"""
    Test cases for salt.states.keystore
"""

import pytest

import salt.states.keystore as keystore
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {keystore: {"__opts__": {"test": False}}}


def test_cert_already_present():
    """
    Test for existing value_present
    """

    cert_return = [
        {
            "valid_until": "August 21 2017",
            "sha1": "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5".encode(
                __salt_system_encoding__
            ),
            "valid_start": "August 22 2012",
            "type": "TrustedCertEntry",
            "alias": "stringhost",
            "expired": True,
        }
    ]
    x509_return = {
        "Not After": "2017-08-21 05:26:54",
        "Subject Hash": "97:95:14:4F",
        "Serial Number": "0D:FA",
        "SHA1 Finger Print": (
            "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5".encode(
                __salt_system_encoding__
            )
        ),
        "SHA-256 Finger Print": "5F:0F:B5:16:65:81:AA:E6:4A:10:1C:15:83:B1:BE:BE:74:E8:14:A9:1E:7A:8A:14:BA:1E:83:5D:78:F6:E9:E7".encode(
            __salt_system_encoding__
        ),
        "MD5 Finger Print": "80:E6:17:AF:78:D8:E4:B8:FB:5F:41:3A:27:1D:CC:F2".encode(
            __salt_system_encoding__
        ),
        "Version": 1,
        "Key Size": 512,
        "Public Key": (
            "-----BEGIN PUBLIC"
            " KEY-----\nMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAJv8ZpB5hEK7qxP9K3v43hUS5fGT4waK\ne7ix4Z4mu5UBv+cw7WSFAt0Vaag0sAbsPzU8Hhsrj/qPABvfB8asUwcCAwEAAQ==\n-----END"
            " PUBLIC KEY-----\n"
        ),
        "Issuer": {
            "C": "JP",
            "organizationName": "Frank4DD",
            "CN": "Frank4DD Web CA",
            "SP": "Tokyo",
            "L": "Chuo-ku",
            "emailAddress": "support@frank4dd.com",
            "OU": "WebCert Support",
        },
        "Issuer Hash": "92:DA:45:6B",
        "Not Before": "2012-08-22 05:26:54",
        "Subject": {
            "C": "JP",
            "SP": "Tokyo",
            "organizationName": "Frank4DD",
            "CN": "www.example.com",
        },
    }

    name = "keystore.jks"
    passphrase = "changeit"
    entries = [
        {
            "alias": "stringhost",
            "certificate": """-----BEGIN CERTIFICATE-----
               MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
               A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
               MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
               YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
               ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
               CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
               ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
               8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
               AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
               8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
               2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
               Hn+GmxZA\n-----END CERTIFICATE-----""",
        }
    ]

    state_return = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "No changes made.\n",
    }

    # with patch.dict(keystore.__opts__, {'test': False}):
    with patch("os.path.exists", return_value=True):
        with patch.dict(
            keystore.__salt__,
            {
                "keystore.list": MagicMock(return_value=cert_return),
                "x509.read_certificate": MagicMock(return_value=x509_return),
            },
        ):
            assert keystore.managed(name, passphrase, entries) == state_return

    with patch("os.path.exists", return_value=True):
        with patch.dict(keystore.__opts__, {"test": True}):
            with patch.dict(
                keystore.__salt__,
                {
                    "keystore.list": MagicMock(return_value=cert_return),
                    "x509.read_certificate": MagicMock(return_value=x509_return),
                },
            ):
                assert keystore.managed(name, passphrase, entries) == state_return


def test_cert_update():
    """
    Test for existing value_present
    """

    cert_return = [
        {
            "valid_until": "August 21 2017",
            "sha1": "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5".encode(
                __salt_system_encoding__
            ),
            "valid_start": "August 22 2012",
            "type": "TrustedCertEntry",
            "alias": "stringhost",
            "expired": True,
        }
    ]
    x509_return = {
        "Not After": "2017-08-21 05:26:54",
        "Subject Hash": "97:95:14:4F",
        "Serial Number": "0D:FA",
        "SHA1 Finger Print": (
            "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D6".encode(
                __salt_system_encoding__
            )
        ),
        "SHA-256 Finger Print": "5F:0F:B5:16:65:81:AA:E6:4A:10:1C:15:83:B1:BE:BE:74:E8:14:A9:1E:7A:8A:14:BA:1E:83:5D:78:F6:E9:E7",
        "MD5 Finger Print": "80:E6:17:AF:78:D8:E4:B8:FB:5F:41:3A:27:1D:CC:F2",
        "Version": 1,
        "Key Size": 512,
        "Public Key": (
            "-----BEGIN PUBLIC"
            " KEY-----\nMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAJv8ZpB5hEK7qxP9K3v43hUS5fGT4waK\ne7ix4Z4mu5UBv+cw7WSFAt0Vaag0sAbsPzU8Hhsrj/qPABvfB8asUwcCAwEAAQ==\n-----END"
            " PUBLIC KEY-----\n"
        ),
        "Issuer": {
            "C": "JP",
            "organizationName": "Frank4DD",
            "CN": "Frank4DD Web CA",
            "SP": "Tokyo",
            "L": "Chuo-ku",
            "emailAddress": "support@frank4dd.com",
            "OU": "WebCert Support",
        },
        "Issuer Hash": "92:DA:45:6B",
        "Not Before": "2012-08-22 05:26:54",
        "Subject": {
            "C": "JP",
            "SP": "Tokyo",
            "organizationName": "Frank4DD",
            "CN": "www.example.com",
        },
    }

    name = "keystore.jks"
    passphrase = "changeit"
    entries = [
        {
            "alias": "stringhost",
            "certificate": """-----BEGIN CERTIFICATE-----
               MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
               A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
               MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
               YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
               ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
               CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
               ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
               8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
               AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
               8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
               2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
               Hn+GmxZA\n-----END CERTIFICATE-----""",
        }
    ]

    test_return = {
        "name": name,
        "changes": {},
        "result": None,
        "comment": "Alias stringhost would have been updated\n",
    }
    state_return = {
        "name": name,
        "changes": {"stringhost": "Updated"},
        "result": True,
        "comment": "Alias stringhost updated.\n",
    }

    with patch.dict(keystore.__opts__, {"test": True}):
        with patch("os.path.exists", return_value=True):
            with patch.dict(
                keystore.__salt__,
                {
                    "keystore.list": MagicMock(return_value=cert_return),
                    "x509.read_certificate": MagicMock(return_value=x509_return),
                },
            ):
                assert keystore.managed(name, passphrase, entries) == test_return

    with patch("os.path.exists", return_value=True):
        with patch.dict(
            keystore.__salt__,
            {
                "keystore.list": MagicMock(return_value=cert_return),
                "x509.read_certificate": MagicMock(return_value=x509_return),
                "keystore.remove": MagicMock(return_value=True),
                "keystore.add": MagicMock(return_value=True),
            },
        ):
            assert keystore.managed(name, passphrase, entries) == state_return


def test_cert_update_no_sha1_fingerprint_in_x509():
    """
    Test for existing value_present
    """

    cert_return = [
        {
            "valid_until": "August 21 2017",
            "sha1": "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5".encode(
                __salt_system_encoding__
            ),
            "valid_start": "August 22 2012",
            "type": "TrustedCertEntry",
            "alias": "stringhost",
            "expired": True,
        }
    ]
    sha1_return = b"07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5"
    x509_return = {
        "Not After": "2017-08-21 05:26:54",
        "Subject Hash": "97:95:14:4F",
        "Serial Number": "0D:FA",
        "SHA-256 Finger Print": "5F:0F:B5:16:65:81:AA:E6:4A:10:1C:15:83:B1:BE:BE:74:E8:14:A9:1E:7A:8A:14:BA:1E:83:5D:78:F6:E9:E7",
        "MD5 Finger Print": "80:E6:17:AF:78:D8:E4:B8:FB:5F:41:3A:27:1D:CC:F2",
        "Version": 1,
        "Key Size": 512,
        "Public Key": (
            "-----BEGIN PUBLIC"
            " KEY-----\nMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAJv8ZpB5hEK7qxP9K3v43hUS5fGT4waK\ne7ix4Z4mu5UBv+cw7WSFAt0Vaag0sAbsPzU8Hhsrj/qPABvfB8asUwcCAwEAAQ==\n-----END"
            " PUBLIC KEY-----\n"
        ),
        "Issuer": {
            "C": "JP",
            "organizationName": "Frank4DD",
            "CN": "Frank4DD Web CA",
            "SP": "Tokyo",
            "L": "Chuo-ku",
            "emailAddress": "support@frank4dd.com",
            "OU": "WebCert Support",
        },
        "Issuer Hash": "92:DA:45:6B",
        "Not Before": "2012-08-22 05:26:54",
        "Subject": {
            "C": "JP",
            "SP": "Tokyo",
            "organizationName": "Frank4DD",
            "CN": "www.example.com",
        },
    }

    name = "keystore.jks"
    passphrase = "changeit"
    entries = [
        {
            "alias": "stringhost",
            "certificate": """-----BEGIN CERTIFICATE-----
               MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
               A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
               MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
               YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
               ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
               CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
               ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
               8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
               AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
               8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
               2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
               Hn+GmxZA\n-----END CERTIFICATE-----""",
        }
    ]

    test_return = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "No changes made.\n",
    }
    with patch("os.path.exists", return_value=True):
        with patch.dict(keystore.__opts__, {"test": True}):
            with patch.dict(
                keystore.__salt__,
                {
                    "keystore.list": MagicMock(return_value=cert_return),
                    "x509.read_certificate": MagicMock(return_value=x509_return),
                    "keystore.get_sha1": MagicMock(return_value=sha1_return),
                },
            ):
                assert keystore.managed(name, passphrase, entries) == test_return


def test_new_file():
    """
    Test for existing value_present
    """
    name = "keystore.jks"
    passphrase = "changeit"
    entries = [
        {
            "alias": "stringhost",
            "certificate": """-----BEGIN CERTIFICATE-----
               MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
               A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
               MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
               YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
               ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
               CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
               ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
               8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
               AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
               8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
               2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
               Hn+GmxZA\n-----END CERTIFICATE-----""",
        }
    ]

    test_return = {
        "name": name,
        "changes": {},
        "result": None,
        "comment": "Alias stringhost would have been added\n",
    }
    state_return = {
        "name": name,
        "changes": {"stringhost": "Added"},
        "result": True,
        "comment": "Alias stringhost added.\n",
    }

    with patch.dict(keystore.__opts__, {"test": True}):
        assert keystore.managed(name, passphrase, entries) == test_return

    with patch("os.path.exists", return_value=False):
        with patch.dict(
            keystore.__salt__,
            {
                "keystore.remove": MagicMock(return_value=True),
                "keystore.add": MagicMock(return_value=True),
            },
        ):
            assert keystore.managed(name, passphrase, entries) == state_return


def test_force_remove():
    """
    Test for existing value_present
    """

    cert_return = [
        {
            "valid_until": "August 21 2017",
            "sha1": "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D5".encode(
                __salt_system_encoding__
            ),
            "valid_start": "August 22 2012",
            "type": "TrustedCertEntry",
            "alias": "oldhost",
            "expired": True,
        }
    ]
    x509_return = {
        "Not After": "2017-08-21 05:26:54",
        "Subject Hash": "97:95:14:4F",
        "Serial Number": "0D:FA",
        "SHA1 Finger Print": (
            "07:1C:B9:4F:0C:C8:51:4D:02:41:24:70:8E:E8:B2:68:7B:D7:D9:D6".encode(
                __salt_system_encoding__
            )
        ),
        "SHA-256 Finger Print": "5F:0F:B5:16:65:81:AA:E6:4A:10:1C:15:83:B1:BE:BE:74:E8:14:A9:1E:7A:8A:14:BA:1E:83:5D:78:F6:E9:E7",
        "MD5 Finger Print": "80:E6:17:AF:78:D8:E4:B8:FB:5F:41:3A:27:1D:CC:F2",
        "Version": 1,
        "Key Size": 512,
        "Public Key": (
            "-----BEGIN PUBLIC"
            " KEY-----\nMFwwDQYJKoZIhvcNAQEBBQADSwAwSAJBAJv8ZpB5hEK7qxP9K3v43hUS5fGT4waK\ne7ix4Z4mu5UBv+cw7WSFAt0Vaag0sAbsPzU8Hhsrj/qPABvfB8asUwcCAwEAAQ==\n-----END"
            " PUBLIC KEY-----\n"
        ),
        "Issuer": {
            "C": "JP",
            "organizationName": "Frank4DD",
            "CN": "Frank4DD Web CA",
            "SP": "Tokyo",
            "L": "Chuo-ku",
            "emailAddress": "support@frank4dd.com",
            "OU": "WebCert Support",
        },
        "Issuer Hash": "92:DA:45:6B",
        "Not Before": "2012-08-22 05:26:54",
        "Subject": {
            "C": "JP",
            "SP": "Tokyo",
            "organizationName": "Frank4DD",
            "CN": "www.example.com",
        },
    }

    name = "keystore.jks"
    passphrase = "changeit"
    entries = [
        {
            "alias": "stringhost",
            "certificate": """-----BEGIN CERTIFICATE-----
               MIICEjCCAXsCAg36MA0GCSqGSIb3DQEBBQUAMIGbMQswCQYDVQQGEwJKUDEOMAwG
               A1UECBMFVG9reW8xEDAOBgNVBAcTB0NodW8ta3UxETAPBgNVBAoTCEZyYW5rNERE
               MRgwFgYDVQQLEw9XZWJDZXJ0IFN1cHBvcnQxGDAWBgNVBAMTD0ZyYW5rNEREIFdl
               YiBDQTEjMCEGCSqGSIb3DQEJARYUc3VwcG9ydEBmcmFuazRkZC5jb20wHhcNMTIw
               ODIyMDUyNjU0WhcNMTcwODIxMDUyNjU0WjBKMQswCQYDVQQGEwJKUDEOMAwGA1UE
               CAwFVG9reW8xETAPBgNVBAoMCEZyYW5rNEREMRgwFgYDVQQDDA93d3cuZXhhbXBs
               ZS5jb20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEAm/xmkHmEQrurE/0re/jeFRLl
               8ZPjBop7uLHhnia7lQG/5zDtZIUC3RVpqDSwBuw/NTweGyuP+o8AG98HxqxTBwID
               AQABMA0GCSqGSIb3DQEBBQUAA4GBABS2TLuBeTPmcaTaUW/LCB2NYOy8GMdzR1mx
               8iBIu2H6/E2tiY3RIevV2OW61qY2/XRQg7YPxx3ffeUugX9F4J/iPnnu1zAxxyBy
               2VguKv4SWjRFoRkIfIlHX0qVviMhSlNy2ioFLy7JcPZb+v3ftDGywUqcBiVDoea0
               Hn+GmxZA\n-----END CERTIFICATE-----""",
        }
    ]

    test_return = {
        "name": name,
        "changes": {},
        "result": None,
        "comment": (
            "Alias stringhost would have been updated\nAlias oldhost would have"
            " been removed"
        ),
    }
    state_return = {
        "name": name,
        "changes": {"oldhost": "Removed", "stringhost": "Updated"},
        "result": True,
        "comment": "Alias stringhost updated.\nAlias oldhost removed.\n",
    }

    with patch.dict(keystore.__opts__, {"test": True}):
        with patch("os.path.exists", return_value=True):
            with patch.dict(
                keystore.__salt__,
                {
                    "keystore.list": MagicMock(return_value=cert_return),
                    "x509.read_certificate": MagicMock(return_value=x509_return),
                },
            ):
                assert (
                    keystore.managed(name, passphrase, entries, force_remove=True)
                    == test_return
                )

    with patch("os.path.exists", return_value=True):
        with patch.dict(
            keystore.__salt__,
            {
                "keystore.list": MagicMock(return_value=cert_return),
                "x509.read_certificate": MagicMock(return_value=x509_return),
                "keystore.remove": MagicMock(return_value=True),
                "keystore.add": MagicMock(return_value=True),
            },
        ):
            assert (
                keystore.managed(name, passphrase, entries, force_remove=True)
                == state_return
            )



import subprocess

import pytest

import salt.version
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES
from tests.support.helpers import VirtualEnv, dedent

pytestmark = [
    pytest.mark.skip_if_binaries_missing(*KNOWN_BINARY_NAMES, check_all=False),
    pytest.mark.requires_network,
]


@pytest.fixture(scope="module")
def _extra_requirements():
    extra_requirements = []
    for name, version in salt.version.dependency_information():
        if name in ["PyYAML", "packaging", "looseversion"]:
            extra_requirements.append(f"{name}=={version}")
    return extra_requirements


@pytest.mark.slow_test
@pytest.mark.parametrize(
    "pip_contraint",
    [
        # Latest pip 18
        "<19.0",
        # Latest pip 19
        "<20.0",
        # Latest pip 20
        "<21.0",
        # Latest pip
        None,
    ],
)
def test_importable_installation_error(_extra_requirements, pip_contraint):
    code = dedent(
        """\
    import sys
    import traceback
    try:
        import salt.states.pip_state
        salt.states.pip_state.InstallationError
    except ImportError as exc:
        traceback.print_exc(file=sys.stdout)
        sys.stdout.flush()
        sys.exit(1)
    except AttributeError as exc:
        traceback.print_exc(file=sys.stdout)
        sys.stdout.flush()
        sys.exit(2)
    except Exception as exc:
        traceback.print_exc(exc, file=sys.stdout)
        sys.stdout.flush()
        sys.exit(3)
    sys.exit(0)
    """
    )
    with VirtualEnv() as venv:
        venv.install(*_extra_requirements)
        if pip_contraint:
            venv.install(f"pip{pip_contraint}")
        try:
            subprocess.check_output([venv.venv_python, "-c", code])
        except subprocess.CalledProcessError as exc:
            if exc.returncode == 1:
                pytest.fail(f"Failed to import pip:\n{exc.output}")
            elif exc.returncode == 2:
                pytest.fail(
                    f"Failed to import InstallationError from pip:\n{exc.output}"
                )
            else:
                pytest.fail(exc.output)



import logging
import os
import textwrap
import threading
import time
from textwrap import dedent

import pytest

import salt.loader
import salt.modules.cmdmod as cmd
import salt.modules.config as config
import salt.modules.grains as grains
import salt.modules.saltutil as saltutil
import salt.modules.state as state_mod
import salt.utils.atomicfile
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.state as state_util
import salt.utils.stringutils

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        state_mod: {
            "__opts__": minion_opts,
            "__salt__": {
                "config.option": config.option,
                "config.get": config.get,
                "saltutil.is_running": saltutil.is_running,
                "grains.get": grains.get,
                "cmd.run": cmd.run,
            },
            "__utils__": {"state.check_result": state_util.check_result},
        },
        config: {
            "__opts__": minion_opts,
        },
        saltutil: {
            "__opts__": minion_opts,
        },
        grains: {
            "__opts__": minion_opts,
        },
    }


def _check_skip(grains):
    if grains["os"] == "SUSE":
        return True
    return False


def test_show_highstate(state, state_testfile_dest_path):
    """
    state.show_highstate
    """
    high = state.show_highstate()
    assert isinstance(high, dict)
    assert str(state_testfile_dest_path) in high
    assert high[str(state_testfile_dest_path)]["__env__"] == "base"


def test_show_lowstate(state):
    """
    state.show_lowstate
    """
    low = state.show_lowstate()
    assert isinstance(low, list)
    for entry in low:
        assert isinstance(entry, dict)


def test_show_states(state):
    """
    state.show_states
    """
    states = state.show_states()
    assert isinstance(states, list)
    for entry in states:
        assert isinstance(entry, str)
    assert states == ["core"]


def test_show_states_missing_sls(state, state_tree):
    """
    Test state.show_states with a sls file
    defined in a top file is missing
    """
    top_sls_contents = """
    base:
      '*':
        - core
        - does-not-exist
    """
    with pytest.helpers.temp_file("top.sls", top_sls_contents, state_tree):
        states = state.show_states()
        assert isinstance(states, list)
        assert states == ["No matching sls found for 'does-not-exist' in env 'base'"]


def test_catch_recurse(state, state_tree):
    """
    state.show_sls used to catch a recursive ref
    """
    sls_contents = """
    mysql:
      service:
        - running
        - require:
          - file: /etc/mysql/my.cnf

    /etc/mysql/my.cnf:
      file:
        - managed
        - source: salt://master.cnf
        - require:
          - service: mysql
    """
    with pytest.helpers.temp_file("recurse-fail.sls", sls_contents, state_tree):
        ret = state.sls("recurse-fail")
        assert ret.failed
        assert (
            'A recursive requisite was found, SLS "recurse-fail" ID "/etc/mysql/my.cnf" ID "mysql"'
            in ret.errors
        )


RECURSE_SLS_ONE = """
snmpd:
  pkg:
    - installed
  service:
    - running
    - require:
      - pkg: snmpd
    - watch:
      - file: /etc/snmp/snmpd.conf

/etc/snmp/snmpd.conf:
  file:
    - managed
    - source: salt://snmpd/snmpd.conf.jinja
    - template: jinja
    - user: root
    - group: root
    - mode: "0600"
    - require:
      - pkg: snmpd
"""
RECURSE_SLS_TWO = """
nagios-nrpe-server:
  pkg:
    - installed
  service:
    - running
    - watch:
      - file: /etc/nagios/nrpe.cfg

/etc/nagios/nrpe.cfg:
  file:
    - managed
    - source: salt://baseserver/nrpe.cfg
    - require:
      - pkg: nagios-nrpe-server
"""


@pytest.mark.parametrize(
    "sls_contents, expected_in_output",
    [(RECURSE_SLS_ONE, "snmpd"), (RECURSE_SLS_TWO, "/etc/nagios/nrpe.cfg")],
    ids=("recurse-scenario-1", "recurse-scenario-2"),
)
def test_no_recurse(state, state_tree, sls_contents, expected_in_output):
    """
    verify that a sls structure is NOT a recursive ref
    """
    with pytest.helpers.temp_file("recurse-ok.sls", sls_contents, state_tree):
        ret = state.show_sls("recurse-ok")
        assert expected_in_output in ret


def test_running_dictionary_consistency(state):
    """
    Test the structure of the running dictionary so we don't change it
    without deprecating/documenting the change
    """
    running_dict_fields = {
        "__id__",
        "__run_num__",
        "__sls__",
        "changes",
        "comment",
        "duration",
        "name",
        "result",
        "start_time",
    }

    sls = state.single("test.succeed_without_changes", name="gndn")
    ret_values_set = set(sls.full_return.keys())
    assert running_dict_fields.issubset(ret_values_set)


def test_running_dictionary_key_sls(state, state_tree):
    """
    Ensure the __sls__ key is either null or a string
    """
    sls1 = state.single("test.succeed_with_changes", name="gndn")
    assert "__sls__" in sls1.full_return
    assert sls1.full_return["__sls__"] is None

    sls_contents = """
    gndn:
      test.succeed_with_changes
    """
    with pytest.helpers.temp_file("gndn.sls", sls_contents, state_tree):
        sls2 = state.sls(mods="gndn")

    for state_return in sls2:
        assert "__sls__" in state_return.full_return
        assert isinstance(state_return.full_return["__sls__"], str)


@pytest.fixture
def requested_sls_key(minion_opts, state_tree):
    if not salt.utils.platform.is_windows():
        sls_contents = """
        count_root_dir_contents:
          cmd.run:
            - name: 'ls -a / | wc -l'
        """
        sls_key = "cmd_|-count_root_dir_contents_|-ls -a / | wc -l_|-run"
    else:
        sls_contents = r"""
        count_root_dir_contents:
          cmd.run:
            - name: 'Get-ChildItem C:\ | Measure-Object | %{$_.Count}'
            - shell: powershell
        """
        sls_key = (
            r"cmd_|-count_root_dir_contents_|-Get-ChildItem C:\ | Measure-Object |"
            r" %{$_.Count}_|-run"
        )
    try:
        with pytest.helpers.temp_file(
            "requested.sls", sls_contents, state_tree
        ) as sls_path:
            yield sls_key
    finally:
        cache_file = os.path.join(minion_opts["cachedir"], "req_state.p")
        if os.path.exists(cache_file):
            os.remove(cache_file)


def test_request(state, requested_sls_key):
    """
    verify sending a state request to the minion(s)
    """
    ret = state.request("requested")
    assert ret[requested_sls_key]["result"] is None


def test_check_request(state, requested_sls_key):
    """
    verify checking a state request sent to the minion(s)
    """
    ret = state.request("requested")
    assert ret[requested_sls_key]["result"] is None

    ret = state.check_request()
    assert ret["default"]["test_run"][requested_sls_key]["result"] is None


def test_clear_request(state, requested_sls_key):
    """
    verify clearing a state request sent to the minion(s)
    """
    ret = state.request("requested")
    assert ret[requested_sls_key]["result"] is None

    ret = state.clear_request()
    assert ret is True


def test_run_request_succeeded(state, requested_sls_key):
    """
    verify running a state request sent to the minion(s)
    """
    ret = state.request("requested")
    assert ret[requested_sls_key]["result"] is None

    ret = state.run_request()
    assert ret[requested_sls_key]["result"] is True


def test_run_request_failed_no_request_staged(state, requested_sls_key):
    """
    verify not running a state request sent to the minion(s)
    """
    ret = state.request("requested")
    assert ret[requested_sls_key]["result"] is None

    ret = state.clear_request()
    assert ret is True

    ret = state.run_request()
    assert ret == {}


def test_issue_1876_syntax_error(state, state_tree, tmp_path):
    """
    verify that we catch the following syntax error::

        /tmp/salttest/issue-1876:

          file:
            - managed
            - source: salt://testfile

          file.append:
            - text: foo

    """
    testfile = tmp_path / "issue-1876.txt"
    sls_contents = """
    {}:
      file:
        - managed
        - source: salt://testfile

      file.append:
        - text: foo
    """.format(
        testfile
    )
    with pytest.helpers.temp_file("issue-1876.sls", sls_contents, state_tree):
        ret = state.sls("issue-1876")
        assert ret.failed
        errmsg = (
            "ID '{}' in SLS 'issue-1876' contains multiple state declarations of the"
            " same type".format(testfile)
        )
        assert errmsg in ret.errors


def test_issue_1879_too_simple_contains_check(state, state_tree, tmp_path):
    testfile = tmp_path / "issue-1979.txt"
    init_sls_contents = """
    {}:
      file:
        - touch
    """.format(
        testfile
    )
    step1_sls_contents = """
    {}:
      file.append:
        - text: |
            # set variable identifying the chroot you work in (used in the prompt below)
            if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
                debian_chroot=$(cat /etc/debian_chroot)
            fi

    """.format(
        testfile
    )
    step2_sls_contents = """
    {}:
      file.append:
        - text: |
            # enable bash completion in interactive shells
            if [ -f /etc/bash_completion ] && ! shopt -oq posix; then
                . /etc/bash_completion
            fi

    """.format(
        testfile
    )

    expected = textwrap.dedent(
        """\
        # set variable identifying the chroot you work in (used in the prompt below)
        if [ -z "$debian_chroot" ] && [ -r /etc/debian_chroot ]; then
            debian_chroot=$(cat /etc/debian_chroot)
        fi
        # enable bash completion in interactive shells
        if [ -f /etc/bash_completion ] && ! shopt -oq posix; then
            . /etc/bash_completion
        fi
        """
    )

    issue_1879_dir = state_tree / "issue-1879"
    with pytest.helpers.temp_file(
        "init.sls", init_sls_contents, issue_1879_dir
    ), pytest.helpers.temp_file(
        "step-1.sls", step1_sls_contents, issue_1879_dir
    ), pytest.helpers.temp_file(
        "step-2.sls", step2_sls_contents, issue_1879_dir
    ):
        # Create the file
        ret = state.sls("issue-1879")
        for staterun in ret:
            assert staterun.result is True

        # The first append
        ret = state.sls("issue-1879.step-1")
        for staterun in ret:
            assert staterun.result is True

        # The second append
        ret = state.sls("issue-1879.step-2")
        for staterun in ret:
            assert staterun.result is True

        # Does it match?
        contents = testfile.read_text()
        assert contents == expected

        # Make sure we don't re-append existing text
        ret = state.sls("issue-1879.step-1")
        for staterun in ret:
            assert staterun.result is True

        ret = state.sls("issue-1879.step-2")
        for staterun in ret:
            assert staterun.result is True

        # Does it match?
        contents = testfile.read_text()
        assert contents == expected


def test_include(state, state_tree, tmp_path):
    testfile_path = tmp_path / "testfile"
    testfile_path.write_text("foo")
    include_test_path = tmp_path / "include-test.txt"
    to_include_test_path = tmp_path / "to-include-test.txt"
    exclude_test_path = tmp_path / "exclude-test.txt"
    to_include_sls_contents = """
    {}:
      file.managed:
        - source: salt://testfile
    """.format(
        to_include_test_path
    )
    include_sls_contents = """
    include:
      - to-include-test

    {}:
      file.managed:
        - source: salt://testfile
    """.format(
        include_test_path
    )
    with pytest.helpers.temp_file(
        "testfile", "foo", state_tree
    ), pytest.helpers.temp_file(
        "to-include-test.sls", to_include_sls_contents, state_tree
    ), pytest.helpers.temp_file(
        "include-test.sls", include_sls_contents, state_tree
    ):
        ret = state.sls("include-test")
        for staterun in ret:
            assert staterun.result is True

    assert include_test_path.exists()
    assert to_include_test_path.exists()
    assert exclude_test_path.exists() is False


def test_exclude(state, state_tree, tmp_path):
    testfile_path = tmp_path / "testfile"
    testfile_path.write_text("foo")
    include_test_path = tmp_path / "include-test.txt"
    to_include_test_path = tmp_path / "to-include-test.txt"
    exclude_test_path = tmp_path / "exclude-test.txt"
    to_include_sls_contents = """
    {}:
      file.managed:
        - source: salt://testfile
    """.format(
        to_include_test_path
    )
    include_sls_contents = """
    include:
      - to-include-test

    {}:
      file.managed:
        - source: salt://testfile
    """.format(
        include_test_path
    )
    exclude_sls_contents = """
    exclude:
      - to-include-test

    include:
      - include-test

    {}:
      file.managed:
        - source: salt://testfile
    """.format(
        exclude_test_path
    )
    with pytest.helpers.temp_file(
        "testfile", "foo", state_tree
    ), pytest.helpers.temp_file(
        "to-include-test.sls", to_include_sls_contents, state_tree
    ), pytest.helpers.temp_file(
        "include-test.sls", include_sls_contents, state_tree
    ), pytest.helpers.temp_file(
        "exclude-test.sls", exclude_sls_contents, state_tree
    ):
        ret = state.sls("exclude-test")
        for staterun in ret:
            assert staterun.result is True

    assert include_test_path.exists()
    assert exclude_test_path.exists()
    assert to_include_test_path.exists() is False


def test_issue_2068_template_str(state, state_tree):
    template_str_no_dot_sls_contents = """
    required_state:
      test:
        - succeed_without_changes

    requiring_state:
      test:
        - succeed_without_changes
        - require:
          - test: required_state
    """
    template_str_sls_contents = """
    required_state: test.succeed_without_changes

    requiring_state:
      test.succeed_without_changes:
        - require:
          - test: required_state
    """
    with pytest.helpers.temp_file(
        "issue-2068-no-dot.sls", template_str_no_dot_sls_contents, state_tree
    ) as template_str_no_dot_path, pytest.helpers.temp_file(
        "issue-2068.sls", template_str_sls_contents, state_tree
    ) as template_str_path:
        # If running this state with state.sls works, so should using state.template_str
        ret = state.sls("issue-2068-no-dot")
        for staterun in ret:
            assert staterun.result is True

        template_str_no_dot_contents = template_str_no_dot_path.read_text()
        ret = state.template_str(template_str_no_dot_contents)
        for staterun in ret:
            assert staterun.result is True

        # Now using state.template
        ret = state.template(str(template_str_no_dot_path))
        for staterun in ret:
            assert staterun.result is True

        # Now the problematic #2068 including dot's
        ret = state.sls("issue-2068")
        for staterun in ret:
            assert staterun.result is True

        template_str_contents = template_str_path.read_text()
        ret = state.template_str(template_str_contents)
        for staterun in ret:
            assert staterun.result is True

        # Now using state.template
        ret = state.template(str(template_str_path))
        for staterun in ret:
            assert staterun.result is True


@pytest.mark.parametrize("item", ("include", "exclude", "extends"))
def test_template_str_invalid_items(state, item):
    TEMPLATE = textwrap.dedent(
        """\
        {}:
          - existing-state

        /tmp/test-template-invalid-items:
          file:
            - managed
            - source: salt://testfile
        """.format(
            item
        )
    )

    ret = state.template_str(TEMPLATE.format(item))
    assert ret.failed
    errmsg = (
        "The '{}' declaration found on '<template-str>' is invalid when "
        "rendering single templates".format(item)
    )
    assert errmsg in ret.errors


@pytest.mark.skip_on_windows(
    reason=(
        "Functional testing this on windows raises unicode errors. "
        "Tested in tests/pytests/integration/modules/state/test_state.py"
    )
)
def test_pydsl(state, state_tree, tmp_path):
    """
    Test the basics of the pydsl
    """
    testfile = tmp_path / "testfile"
    sls_contents = """
    #!pydsl

    state("{}").file("touch")
    """.format(
        testfile
    )
    with pytest.helpers.temp_file("pydsl.sls", sls_contents, state_tree):
        ret = state.sls("pydsl")
        for staterun in ret:
            assert staterun.result is True
        assert testfile.exists()


def test_issues_7905_and_8174_sls_syntax_error(state, state_tree):
    """
    Call sls file with yaml syntax error.

    Ensure theses errors are detected and presented to the user without
    stack traces.
    """
    badlist_1_sls_contents = """
    # Missing " " between "-" and "foo" or "name"
    A:
      cmd.run:
        -name: echo foo
        -foo:
          - bar
    """
    badlist_2_sls_contents = """
    # C should fail with bad list error message
    B:
      # ok
      file.exist:
        - name: /foo/bar/foobar
    # ok
    /foo/bar/foobar:
      file.exist

    # nok
    C:
      /foo/bar/foobar:
        file.exist
    """
    with pytest.helpers.temp_file(
        "badlist1.sls", badlist_1_sls_contents, state_tree
    ), pytest.helpers.temp_file("badlist2.sls", badlist_2_sls_contents, state_tree):
        ret = state.sls("badlist1")
        assert ret.failed
        assert ret.errors == ["State 'A' in SLS 'badlist1' is not formed as a list"]

        ret = state.sls("badlist2")
        assert ret.failed
        assert ret.errors == ["State 'C' in SLS 'badlist2' is not formed as a list"]


def test_retry_option(state, state_tree):
    """
    test the retry option on a simple state with defaults
    ensure comment is as expected
    ensure state duration is greater than configured the passed (interval * attempts)
    """
    sls_contents = """
    file_test:
      file.exists:
        - name: /path/to/a/non-existent/file.txt
        - retry:
            until: True
            attempts: 3
            interval: 1
            splay: 0
    """
    expected_comment = (
        'Attempt 1: Returned a result of "False", with the following '
        'comment: "Specified path /path/to/a/non-existent/file.txt does not exist"'
    )
    with pytest.helpers.temp_file("retry.sls", sls_contents, state_tree):
        ret = state.sls("retry")
        for state_return in ret:
            assert state_return.result is False
            assert expected_comment in state_return.comment
            assert state_return.full_return["duration"] >= 3


def test_retry_option_is_true(state, state_tree):
    """
    test the retry: True on a simple state with defaults
    ensure comment is as expected
    ensure state duration is greater than configured the passed (interval * attempts)
    """
    sls_contents = """
    file_test:
      file.exists:
        - name: /path/to/a/non-existent/file.txt
        - retry: True
    """
    expected_comment = (
        'Attempt 1: Returned a result of "False", with the following '
        'comment: "Specified path /path/to/a/non-existent/file.txt does not exist"'
    )
    with pytest.helpers.temp_file("retry.sls", sls_contents, state_tree):
        ret = state.sls("retry")
        for state_return in ret:
            assert state_return.result is False
            assert expected_comment in state_return.comment
            assert state_return.full_return["duration"] >= 3


@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
def test_retry_option_success(state, state_tree, tmp_path):
    """
    test a state with the retry option that should return True immediately (i.e. no retries)
    """
    testfile = tmp_path / "testfile"
    testfile.touch()
    sls_contents = """
    file_test:
      file.exists:
        - name: {}
        - retry:
            until: True
            attempts: 5
            interval: 2
            splay: 0
    """.format(
        testfile
    )
    duration = 4
    if salt.utils.platform.spawning_platform():
        duration = 16

    with pytest.helpers.temp_file("retry.sls", sls_contents, state_tree):
        ret = state.sls("retry")
        for state_return in ret:
            assert state_return.result is True
            assert state_return.full_return["duration"] < duration
            # It should not take 2 attempts
            assert "Attempt 2" not in state_return.comment


@pytest.mark.skip_on_windows(
    reason="Skipped until parallel states can be fixed on Windows"
)
def test_retry_option_success_parallel(state, state_tree, tmp_path):
    """
    test a state with the retry option that should return True immediately (i.e. no retries)
    """
    testfile = tmp_path / "testfile"
    testfile.touch()
    sls_contents = """
    file_test:
      file.exists:
        - name: {}
        - parallel: True
        - retry:
            until: True
            attempts: 5
            interval: 2
            splay: 0
    """.format(
        testfile
    )
    duration = 4
    if salt.utils.platform.spawning_platform():
        duration = 30
        # mac needs some more time to do its makeup
        if salt.utils.platform.is_darwin():
            duration += 15

    with pytest.helpers.temp_file("retry.sls", sls_contents, state_tree):
        ret = state.sls(
            "retry", __pub_jid="1"
        )  # Because these run in parallel we need a fake JID

        for state_return in ret:
            assert state_return.result is True
            assert state_return.full_return["duration"] < duration
            # It should not take 2 attempts
            assert "Attempt 2" not in state_return.comment


def test_retry_option_eventual_success(state, state_tree, tmp_path):
    """
    test a state with the retry option that should return True, eventually
    """
    testfile1 = tmp_path / "testfile-1"
    testfile2 = tmp_path / "testfile-2"

    def create_testfile(testfile1, testfile2):
        while True:
            if testfile1.exists():
                break
        time.sleep(2)
        testfile2.touch()

    thread = threading.Thread(target=create_testfile, args=(testfile1, testfile2))
    sls_contents = """
    file_test_a:
      file.managed:
        - name: {}
        - content: 'a'

    file_test:
      file.exists:
        - name: {}
        - retry:
            until: True
            attempts: 5
            interval: 2
            splay: 0
        - require:
          - file_test_a
    """.format(
        testfile1, testfile2
    )
    with pytest.helpers.temp_file("retry.sls", sls_contents, state_tree):
        thread.start()
        ret = state.sls("retry")
        for state_return in ret:
            assert state_return.result is True
            assert state_return.full_return["duration"] > 4
            # It should not take 5 attempts
            assert "Attempt 5" not in state_return.comment


@pytest.mark.skip_on_windows(
    reason="Skipped until parallel states can be fixed on Windows"
)
def test_retry_option_eventual_success_parallel(state, state_tree, tmp_path):
    """
    test a state with the retry option that should return True, eventually
    """
    testfile1 = tmp_path / "testfile-1"
    testfile2 = tmp_path / "testfile-2"

    def create_testfile(testfile1, testfile2):
        while True:
            if testfile1.exists():
                break
        time.sleep(2)
        testfile2.touch()

    thread = threading.Thread(target=create_testfile, args=(testfile1, testfile2))
    sls_contents = """
    file_test_a:
      file.managed:
        - name: {}
        - content: 'a'

    file_test:
      file.exists:
        - name: {}
        - retry:
            until: True
            attempts: 5
            interval: 2
            splay: 0
        - parallel: True
        - require:
          - file_test_a
    """.format(
        testfile1, testfile2
    )
    with pytest.helpers.temp_file("retry.sls", sls_contents, state_tree):
        thread.start()
        ret = state.sls(
            "retry", __pub_jid="1"
        )  # Because these run in parallel we need a fake JID
        for state_return in ret:
            log.debug("=== state_return %s ===", state_return)
            assert state_return.result is True
            assert state_return.full_return["duration"] > 4
            # It should not take 5 attempts
            assert "Attempt 5" not in state_return.comment


def test_state_non_base_environment(state, state_tree_prod, tmp_path):
    """
    test state.sls with saltenv using a nonbase environment
    with a salt source
    """
    testfile = tmp_path / "testfile"
    sls_contents = """
    {}:
      file.managed:
        - content: foo
    """.format(
        testfile
    )
    with pytest.helpers.temp_file("non-base-env.sls", sls_contents, state_tree_prod):
        ret = state.sls("non-base-env", saltenv="prod")
        for state_return in ret:
            assert state_return.result is True
        assert testfile.exists()


@pytest.mark.skip_on_windows(
    reason="Skipped until parallel states can be fixed on Windows"
)
def test_parallel_state_with_long_tag(state, state_tree):
    """
    This tests the case where the state being executed has a long ID dec or
    name and states are being run in parallel. The filenames used for the
    parallel state cache were previously based on the tag for each chunk,
    and longer ID decs or name params can cause the cache file to be longer
    than the operating system's max file name length. To counter this we
    instead generate a SHA1 hash of the chunk's tag to use as the cache
    filename. This test will ensure that long tags don't cause caching
    failures.

    See https://github.com/saltstack/salt/issues/49738 for more info.
    """
    short_command = "helloworld"
    long_command = short_command * 25
    sls_contents = """
    test_cmd_short:
      cmd.run:
        - name: {}
        - parallel: True

    test_cmd_long:
      cmd.run:
        - name: {}
        - parallel: True
    """.format(
        short_command, long_command
    )
    with pytest.helpers.temp_file("issue-49738.sls", sls_contents, state_tree):
        ret = state.sls(
            "issue-49738",
            __pub_jid="1",  # Because these run in parallel we need a fake JID
        )

    comments = sorted(x.comment for x in ret)
    expected = sorted(f'Command "{x}" run' for x in (short_command, long_command))
    assert comments == expected, f"{comments} != {expected}"


@pytest.mark.skip_on_darwin(reason="Test is broken on macosx")
@pytest.mark.skip_on_windows(
    reason=(
        "Functional testing this on windows raises unicode errors. "
        "Tested in tests/pytests/integration/modules/state/test_state.py"
    )
)
def test_state_sls_unicode_characters(state, state_tree):
    """
    test state.sls when state file contains non-ascii characters
    """
    sls_contents = """
    echo1:
      cmd.run:
        - name: "echo 'This is Æ test!'"
    """
    with pytest.helpers.temp_file("issue-46672.sls", sls_contents, state_tree):
        ret = state.sls("issue-46672")
        expected = "cmd_|-echo1_|-echo 'This is Æ test!'_|-run"
        assert expected in ret


def test_state_sls_integer_name(state, state_tree):
    """
    This tests the case where the state file is named
    only with integers
    """
    sls_contents = """
    always-passes:
      test.succeed_without_changes
    """
    state_id = "test_|-always-passes_|-always-passes_|-succeed_without_changes"
    with pytest.helpers.temp_file("12345.sls", sls_contents, state_tree):
        ret = state.sls("12345")
        assert state_id in ret
        for state_return in ret:
            assert state_return.result is True
            assert "Success!" in state_return.comment

        ret = state.sls(mods=12345)
        assert state_id in ret
        for state_return in ret:
            assert state_return.result is True
            assert "Success!" in state_return.comment


def test_state_sls_lazyloader_allows_recursion(state, state_tree):
    """
    This tests that referencing dunders like __salt__ work
    context: https://github.com/saltstack/salt/pull/51499
    """
    sls_contents = """
    {% if 'nonexistent_module.function' in salt %}
    {% do salt.log.warning("Module is available") %}
    {% endif %}
    always-passes:
      test.succeed_without_changes:
        - name: foo
    """
    state_id = "test_|-always-passes_|-foo_|-succeed_without_changes"
    with pytest.helpers.temp_file("issue-51499.sls", sls_contents, state_tree):
        ret = state.sls("issue-51499")
        assert state_id in ret
        for state_return in ret:
            assert state_return.result is True
            assert "Success!" in state_return.comment


def test_issue_62264_requisite_not_found(state, state_tree):
    """
    This tests that the proper state module is referenced for _in requisites
    when no explicit state module is given.
    Context: https://github.com/saltstack/salt/pull/62264
    """
    sls_contents = """
    stuff:
      cmd.run:
        - name: echo hello

    thing_test:
      cmd.run:
        - name: echo world
        - require_in:
          - /stuff/*
          - test: service_running

    service_running:
      test.succeed_without_changes:
        - require:
          - cmd: stuff
    """
    with pytest.helpers.temp_file("issue-62264.sls", sls_contents, state_tree):
        ret = state.sls("issue-62264")
        for state_return in ret:
            assert state_return.result is True
            assert "The following requisites were not found" not in state_return.comment


def test_state_sls_defaults(state, state_tree):
    """ """
    json_contents = """
    {
        "users": {
            "root": 1
        }
    }
    """
    sls_contents = """
    {% set test = salt['defaults.get']('core:users:root') %}

    echo {{ test }}:
      cmd.run
    """
    state_id = "cmd_|-echo 1_|-echo 1_|-run"
    core_dir = state_tree / "core"
    with pytest.helpers.temp_file(
        core_dir / "defaults.json", json_contents, state_tree
    ):
        with pytest.helpers.temp_file(core_dir / "test.sls", sls_contents, state_tree):
            ret = state.sls("core.test")
            assert state_id in ret
            for state_return in ret:
                assert state_return.result is True
                assert "echo 1" in state_return.comment


def test_state_sls_mock_ret(state_tree):
    """
    test state.sls when mock=True is passed
    """
    sls_contents = """
    echo1:
      cmd.run:
        - name: "echo 'This is a test!'"
    """
    with pytest.helpers.temp_file("mock.sls", sls_contents, state_tree):
        ret = state_mod.sls("mock", mock=True)
        assert (
            ret["cmd_|-echo1_|-echo 'This is a test!'_|-run"]["comment"]
            == "Not called, mocked"
        )


@pytest.fixture
def _state_requires_env(loaders, state_tree):
    mod_contents = dedent(
        r"""
        def test_it(name):
            return {
                "name": name,
                "result": __env__ == "base",
                "comment": "",
                "changes": {},
            }
        """
    )
    sls = "test_spawning"
    sls_contents = dedent(
        """
        This should not fail on spawning platforms:
          requires_env.test_it:
            - name: foo
            - parallel: true
        """
    )
    with pytest.helpers.temp_file(
        f"{sls}.sls", sls_contents, state_tree
    ), pytest.helpers.temp_file("_states/requires_env.py", mod_contents, state_tree):
        res = loaders.modules.saltutil.sync_states()
        assert "states.requires_env" in res
        yield sls


def test_state_apply_parallel_spawning_with_global_dunders(state, _state_requires_env):
    """
    Ensure state modules called via `parallel: true` have access to injected
    global dunders like `__env__`.
    """
    ret = state.apply(_state_requires_env)
    assert (
        ret[
            "requires_env_|-This should not fail on spawning platforms_|-foo_|-test_it"
        ]["result"]
        is True
    )


@pytest.fixture
def _state_unpicklable_ctx(loaders, state_tree):
    mod_contents = dedent(
        r"""
        import threading

        class Unpicklable:
            def __init__(self):
                self._lock = threading.RLock()

        def test_it():
            __context__["booh"] = Unpicklable()
        """
    )
    sls = "test_spawning_unpicklable"
    sls_contents = dedent(
        r"""
        {%- do salt["unpicklable.test_it"]() %}

        This should not fail on spawning platforms:
          test.nop:
            - name: foo
            - parallel: true
        """
    )
    with pytest.helpers.temp_file(
        f"{sls}.sls", sls_contents, state_tree
    ), pytest.helpers.temp_file("_modules/unpicklable.py", mod_contents, state_tree):
        res = loaders.modules.saltutil.sync_modules()
        assert "modules.unpicklable" in res
        yield sls


@pytest.mark.skip_unless_on_spawning_platform(
    reason="Pickling is only relevant on spawning platforms"
)
def test_state_apply_parallel_spawning_with_unpicklable_context(
    state, _state_unpicklable_ctx
):
    """
    Ensure that if the __context__ dictionary contains unpicklable objects,
    they are filtered out instead of causing a crash.
    """
    ret = state.apply(_state_unpicklable_ctx)
    assert (
        ret["test_|-This should not fail on spawning platforms_|-foo_|-nop"]["result"]
        is True
    )



import time

import pytest

from tests.integration.cloud.helpers.cloud_test_base import TIMEOUT, CloudTest


class VultrTest(CloudTest):
    """
    Integration tests for the Vultr cloud provider in Salt-Cloud
    """

    PROVIDER = "vultr"
    REQUIRED_PROVIDER_CONFIG_ITEMS = ("api_key", "ssh_key_file", "ssh_key_name")

    def test_list_images(self):
        """
        Tests the return of running the --list-images command for Vultr
        """
        image_list = self.run_cloud(f"--list-images {self.PROVIDER}")

        self.assertIn("Debian 10 x64 (buster)", [i.strip() for i in image_list])

    def test_list_locations(self):
        """
        Tests the return of running the --list-locations command for Vultr
        """
        location_list = self.run_cloud(f"--list-locations {self.PROVIDER}")
        self.assertIn("New Jersey", [i.strip() for i in location_list])

    def test_list_sizes(self):
        """
        Tests the return of running the --list-sizes command for Vultr
        """
        size_list = self.run_cloud(f"--list-sizes {self.PROVIDER}")
        self.assertIn(
            "2048 MB RAM,55 GB SSD,2.00 TB BW", [i.strip() for i in size_list]
        )

    # Commented for now, Vultr driver does not yet support key management
    #    def test_key_management(self):
    #        '''
    #        Test key management
    #        '''
    #        pub = 'ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAAAQQDDHr/jh2Jy4yALcK4JyWbVkPRaWmhck3IgCoeOO3z1e2dBowLh64QAM+Qb72pxekALga2oi4GvT+TlWNhzPH4V example'
    #        finger_print = '3b:16:bf:e4:8b:00:8b:b8:59:8c:a9:d3:f0:19:45:fa'
    #
    #        _key = self.run_cloud('-f create_key {0} name="MyPubKey" public_key="{1}"'.format(self.PROVIDER, pub))
    #
    #        # Upload public key
    #        self.assertIn(
    #            finger_print,
    #            [i.strip() for i in _key]
    #        )
    #
    #        try:
    #            # List all keys
    #            list_keypairs = self.run_cloud('-f list_keypairs {0}'.format(self.PROVIDER))
    #
    #            self.assertIn(
    #                finger_print,
    #                [i.strip() for i in list_keypairs]
    #            )
    #
    #            # List key
    #            show_keypair = self.run_cloud('-f show_keypair {0} keyname={1}'.format(self.PROVIDER, 'MyPubKey'))
    #
    #            self.assertIn(
    #                finger_print,
    #                [i.strip() for i in show_keypair]
    #            )
    #        except AssertionError:
    #            # Delete the public key if the above assertions fail
    #            self.run_cloud('-f remove_key {0} id={1}'.format(self.PROVIDER, finger_print))
    #            raise
    #
    #        # Delete public key
    #        self.assertTrue(self.run_cloud('-f remove_key {0} id={1}'.format(self.PROVIDER, finger_print)))

    @pytest.mark.skip(reason="Skipped temporarily")
    def test_instance(self):
        """
        Test creating an instance on Vultr
        """
        # check if instance with salt installed returned
        ret_val = self.run_cloud(
            f"-p vultr-test {self.instance_name}", timeout=TIMEOUT + 300
        )
        self.assertInstanceExists(ret_val)

        # Vultr won't let us delete an instance less than 5 minutes old.
        time.sleep(300)
        self.assertDestroyInstance()



"""
Manage the password database on BSD systems

.. important::
    If you feel that Salt should be using this module to manage passwords on a
    minion, and it is using a different module (or gives an error similar to
    *'shadow.info' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError, SaltInvocationError

try:
    import pwd
except ImportError:
    pass

try:
    import salt.utils.pycrypto

    HAS_CRYPT = True
except ImportError:
    HAS_CRYPT = False

# Define the module's virtual name
__virtualname__ = "shadow"


def __virtual__():
    if "BSD" in __grains__.get("os", ""):
        return __virtualname__
    return (
        False,
        "The bsd_shadow execution module cannot be loaded: "
        "only available on BSD family systems.",
    )


def default_hash():
    """
    Returns the default hash used for unset passwords

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.default_hash
    """
    return "*" if __grains__["os"].lower() == "freebsd" else "*************"


def gen_password(password, crypt_salt=None, algorithm="sha512"):
    """
    Generate hashed password

    .. note::

        When called this function is called directly via remote-execution,
        the password argument may be displayed in the system's process list.
        This may be a security risk on certain systems.

    password
        Plaintext password to be hashed.

    crypt_salt
        Crpytographic salt. If not given, a random 8-character salt will be
        generated.

    algorithm
        The following hash algorithms are supported:

        * md5
        * blowfish (not in mainline glibc, only available in distros that add it)
        * sha256
        * sha512 (default)

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.gen_password 'I_am_password'
        salt '*' shadow.gen_password 'I_am_password' crypt_salt='I_am_salt' algorithm=sha256
    """
    if not HAS_CRYPT:
        raise CommandExecutionError(
            "gen_password is not available on this operating system "
            'because the "crypt" python module is not available.'
        )
    return salt.utils.pycrypto.gen_hash(crypt_salt, password, algorithm)


def info(name):
    """
    Return information for the specified user

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.info someuser
    """
    try:
        data = pwd.getpwnam(name)
        ret = {"name": data.pw_name, "passwd": data.pw_passwd}
    except KeyError:
        return {"name": "", "passwd": ""}

    if not isinstance(name, str):
        name = str(name)
    if ":" in name:
        raise SaltInvocationError(f"Invalid username '{name}'")

    if __salt__["cmd.has_exec"]("pw"):
        change, expire = __salt__["cmd.run_stdout"](
            ["pw", "user", "show", name], python_shell=False
        ).split(":")[5:7]
    elif __grains__["kernel"] in ("NetBSD", "OpenBSD"):
        try:
            with salt.utils.files.fopen("/etc/master.passwd", "r") as fp_:
                for line in fp_:
                    line = salt.utils.stringutils.to_unicode(line)
                    if line.startswith(f"{name}:"):
                        key = line.split(":")
                        change, expire = key[5:7]
                        ret["passwd"] = str(key[1])
                        break
        except OSError:
            change = expire = None
    else:
        change = expire = None

    try:
        ret["change"] = int(change)
    except ValueError:
        pass

    try:
        ret["expire"] = int(expire)
    except ValueError:
        pass

    return ret


def set_change(name, change):
    """
    Sets the time at which the password expires (in seconds since the UNIX
    epoch). See ``man 8 usermod`` on NetBSD and OpenBSD or ``man 8 pw`` on
    FreeBSD.

    A value of ``0`` sets the password to never expire.

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_change username 1419980400
    """
    pre_info = info(name)
    if change == pre_info["change"]:
        return True
    if __grains__["kernel"] == "FreeBSD":
        cmd = ["pw", "user", "mod", name, "-f", change]
    else:
        cmd = ["usermod", "-f", change, name]
    __salt__["cmd.run"](cmd, python_shell=False)
    post_info = info(name)
    if post_info["change"] != pre_info["change"]:
        return post_info["change"] == change


def set_expire(name, expire):
    """
    Sets the time at which the account expires (in seconds since the UNIX
    epoch). See ``man 8 usermod`` on NetBSD and OpenBSD or ``man 8 pw`` on
    FreeBSD.

    A value of ``0`` sets the account to never expire.

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_expire username 1419980400
    """
    pre_info = info(name)
    if expire == pre_info["expire"]:
        return True
    if __grains__["kernel"] == "FreeBSD":
        cmd = ["pw", "user", "mod", name, "-e", expire]
    else:
        cmd = ["usermod", "-e", expire, name]
    __salt__["cmd.run"](cmd, python_shell=False)
    post_info = info(name)
    if post_info["expire"] != pre_info["expire"]:
        return post_info["expire"] == expire


def del_password(name):
    """
    .. versionadded:: 2015.8.2

    Delete the password from name user

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.del_password username
    """
    cmd = f"pw user mod {name} -w none"
    __salt__["cmd.run"](cmd, python_shell=False, output_loglevel="quiet")
    uinfo = info(name)
    return not uinfo["passwd"]


def set_password(name, password):
    """
    Set the password for a named user. The password must be a properly defined
    hash. A password hash can be generated with :py:func:`gen_password`.

    It is important to make sure that a supported cipher is used.

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_password someuser '$1$UYCIxa628.9qXjpQCjM4a..'
    """
    if __grains__.get("os", "") == "FreeBSD":
        cmd = ["pw", "user", "mod", name, "-H", "0"]
        stdin = password
    else:
        cmd = ["usermod", "-p", password, name]
        stdin = None
    __salt__["cmd.run"](cmd, stdin=stdin, output_loglevel="quiet", python_shell=False)
    return info(name)["passwd"] == password



"""
Installation of Cabal Packages
==============================

.. versionadded:: 2015.8.0

These states manage the installed packages for Haskell using
cabal. Note that cabal-install must be installed for these states to
be available, so cabal states should include a requisite to a
pkg.installed state for the package which provides cabal
(``cabal-install`` in case of Debian based distributions). Example::

.. code-block:: yaml

   cabal-install:
     pkg.installed

   ShellCheck:
     cabal.installed:
       - require:
         - pkg: cabal-install

"""

import salt.utils.path
from salt.exceptions import CommandExecutionError, CommandNotFoundError


def __virtual__():
    """
    Only work when cabal-install is installed.
    """
    if (salt.utils.path.which("cabal") is not None) and (
        salt.utils.path.which("ghc-pkg") is not None
    ):
        return True
    return (False, "cabal or ghc-pkg commands not found")


def _parse_pkg_string(pkg):
    """
    Parse pkg string and return a tuple of package name, separator, and
    package version.

    Cabal support install package with following format:

    * foo-1.0
    * foo < 1.2
    * foo > 1.3

    For the sake of simplicity only the first form is supported,
    support for other forms can be added later.
    """
    pkg_name, separator, pkg_ver = pkg.partition("-")
    return (pkg_name.strip(), separator, pkg_ver.strip())


def installed(name, pkgs=None, user=None, install_global=False, env=None):
    """
    Verify that the given package is installed and is at the correct version
    (if specified).

    .. code-block:: yaml

        ShellCheck-0.3.5:
          cabal:
            - installed:

    name
        The package to install
    user
        The user to run cabal install with
    install_global
        Install package globally instead of locally
    env
        A list of environment variables to be set prior to execution. The
        format is the same as the :py:func:`cmd.run <salt.states.cmd.run>`.
        state function.
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    try:
        call = __salt__["cabal.update"](user=user, env=env)
    except (CommandNotFoundError, CommandExecutionError) as err:
        ret["result"] = False
        ret["comment"] = f"Could not run cabal update {err}"
        return ret

    if pkgs is not None:
        pkg_list = pkgs
    else:
        pkg_list = [name]

    try:
        installed_pkgs = __salt__["cabal.list"](user=user, installed=True, env=env)
    except (CommandNotFoundError, CommandExecutionError) as err:
        ret["result"] = False
        ret["comment"] = f"Error looking up '{name}': {err}"
        return ret

    pkgs_satisfied = []
    pkgs_to_install = []

    for pkg in pkg_list:
        pkg_name, _, pkg_ver = _parse_pkg_string(pkg)

        if pkg_name not in installed_pkgs:
            pkgs_to_install.append(pkg)
        else:
            if pkg_ver:  # version is specified
                if installed_pkgs[pkg_name] != pkg_ver:
                    pkgs_to_install.append(pkg)
                else:
                    pkgs_satisfied.append(pkg)
            else:
                pkgs_satisfied.append(pkg)

    if __opts__["test"]:
        ret["result"] = None

        comment_msg = []

        if pkgs_to_install:
            comment_msg.append(
                "Packages(s) '{}' are set to be installed".format(
                    ", ".join(pkgs_to_install)
                )
            )

        if pkgs_satisfied:
            comment_msg.append(
                "Packages(s) '{}' satisfied by {}".format(
                    ", ".join(pkg_list), ", ".join(pkgs_satisfied)
                )
            )

        ret["comment"] = ". ".join(comment_msg)
        return ret

    if not pkgs_to_install:
        ret["result"] = True
        ret["comment"] = "Packages(s) '{}' satisfied by {}".format(
            ", ".join(pkg_list), ", ".join(pkgs_satisfied)
        )

        return ret

    try:
        call = __salt__["cabal.install"](
            pkgs=pkg_list, user=user, install_global=install_global, env=env
        )
    except (CommandNotFoundError, CommandExecutionError) as err:
        ret["result"] = False
        ret["comment"] = "Error installing '{}': {}".format(", ".join(pkg_list), err)
        return ret

    if call and isinstance(call, dict):
        ret["result"] = True
        ret["changes"] = {"old": [], "new": pkgs_to_install}
        ret["comment"] = "Packages(s) '{}' successfully installed".format(
            ", ".join(pkgs_to_install)
        )
    else:
        ret["result"] = False
        ret["comment"] = "Could not install packages(s) '{}'".format(
            ", ".join(pkg_list)
        )

    return ret


def removed(name, user=None, env=None):
    """
    Verify that given package is not installed.
    """

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    try:
        installed_pkgs = __salt__["cabal.list"](user=user, installed=True, env=env)
    except (CommandNotFoundError, CommandExecutionError) as err:
        ret["result"] = False
        ret["comment"] = f"Error looking up '{name}': {err}"

    if name not in installed_pkgs:
        ret["result"] = True
        ret["comment"] = f"Package '{name}' is not installed"
        return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Package '{name}' is set to be removed"
        return ret

    if __salt__["cabal.uninstall"](pkg=name, user=user, env=env):
        ret["result"] = True
        ret["changes"][name] = "Removed"
        ret["comment"] = f"Package '{name}' was successfully removed"
    else:
        ret["result"] = False
        ret["comment"] = f"Error removing package '{name}'"

    return ret



import pytest

from tests.support.mock import MagicMock, patch


@pytest.fixture
def app(app):
    app.wsgi_application.config["global"]["tools.cors_tool.on"] = True
    return app


async def test_option_request(http_client):
    with patch(
        "salt.netapi.rest_cherrypy.app.cherrypy.session", MagicMock(), create=True
    ):
        response = await http_client.fetch(
            "/", method="OPTIONS", headers={"Origin": "https://domain.com"}
        )
    assert response.code == 200
    assert response.headers["Access-Control-Allow-Origin"] == "https://domain.com"



import pytest


@pytest.fixture(scope="module")
def salt_proxy(salt_master, salt_proxy_factory):
    """
    A running salt-proxy fixture
    """
    assert salt_master.is_running()
    with salt_proxy_factory.started():
        yield salt_proxy_factory


@pytest.fixture(scope="module")
def deltaproxy_parallel_startup():
    yield from [True, False]


@pytest.fixture(
    scope="module",
    params=[True, False],
    ids=["parallel_startup=True", "parallel_startup=False"],
)
def deltaproxy_pillar_tree(request, salt_master, salt_delta_proxy_factory):
    """
    Create the pillar files for controlproxy and two dummy proxy minions
    """
    (
        proxy_one,
        proxy_two,
        proxy_three,
        proxy_four,
    ) = pytest.helpers.proxy.delta_proxy_minion_ids()

    top_file = """
    base:
      {control}:
        - controlproxy
      {one}:
        - {one}
      {two}:
        - {two}
      {three}:
        - {three}
      {four}:
        - {four}
    """.format(
        control=salt_delta_proxy_factory.id,
        one=proxy_one,
        two=proxy_two,
        three=proxy_three,
        four=proxy_four,
    )
    controlproxy_pillar_file = """
    proxy:
        proxytype: deltaproxy
        parallel_startup: {}
        ids:
          - {}
          - {}
          - {}
          - {}
    """.format(
        request.param,
        proxy_one,
        proxy_two,
        proxy_three,
        proxy_four,
    )

    dummy_proxy_pillar_file = """
    proxy:
      proxytype: dummy
    """

    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    controlproxy_tempfile = salt_master.pillar_tree.base.temp_file(
        "controlproxy.sls", controlproxy_pillar_file
    )
    dummy_proxy_one_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_one}.sls", dummy_proxy_pillar_file
    )
    dummy_proxy_two_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_two}.sls", dummy_proxy_pillar_file
    )
    dummy_proxy_three_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_three}.sls", dummy_proxy_pillar_file
    )
    dummy_proxy_four_tempfile = salt_master.pillar_tree.base.temp_file(
        f"{proxy_four}.sls", dummy_proxy_pillar_file
    )
    with (
        top_tempfile
    ), (
        controlproxy_tempfile
    ), (
        dummy_proxy_one_tempfile
    ), dummy_proxy_two_tempfile, dummy_proxy_three_tempfile, dummy_proxy_four_tempfile:
        yield


@pytest.fixture(scope="module")
def salt_delta_proxy(salt_master, salt_delta_proxy_factory, deltaproxy_pillar_tree):
    """
    A running salt-proxy fixture
    """
    assert salt_master.is_running()
    with salt_delta_proxy_factory.started():
        yield salt_delta_proxy_factory



"""
    :codeauthor: :email:`Alexandru Bleotu <alexandru.bleotu@morganstanley.com>`

    Tests functions in salt.utils.vsan
"""

import logging

import pytest

import salt.utils.pbm
from salt.exceptions import (
    VMwareApiError,
    VMwareObjectRetrievalError,
    VMwareRuntimeError,
)
from tests.support.mock import MagicMock, PropertyMock, patch
from tests.support.unit import TestCase

try:
    from pyVmomi import pbm, vim, vmodl  # pylint: disable=no-name-in-module

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


# Get Logging Started
log = logging.getLogger(__name__)


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class GetProfileManagerTestCase(TestCase):
    """Tests for salt.utils.pbm.get_profile_manager"""

    def setUp(self):
        self.mock_si = MagicMock()
        self.mock_stub = MagicMock()
        self.mock_prof_mgr = MagicMock()
        self.mock_content = MagicMock()
        self.mock_pbm_si = MagicMock(
            RetrieveContent=MagicMock(return_value=self.mock_content)
        )
        type(self.mock_content).profileManager = PropertyMock(
            return_value=self.mock_prof_mgr
        )
        patches = (
            (
                "salt.utils.vmware.get_new_service_instance_stub",
                MagicMock(return_value=self.mock_stub),
            ),
            (
                "salt.utils.pbm.pbm.ServiceInstance",
                MagicMock(return_value=self.mock_pbm_si),
            ),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

    def tearDown(self):
        for attr in (
            "mock_si",
            "mock_stub",
            "mock_content",
            "mock_pbm_si",
            "mock_prof_mgr",
        ):
            delattr(self, attr)

    def test_get_new_service_stub(self):
        mock_get_new_service_stub = MagicMock()
        with patch(
            "salt.utils.vmware.get_new_service_instance_stub", mock_get_new_service_stub
        ):
            salt.utils.pbm.get_profile_manager(self.mock_si)
        mock_get_new_service_stub.assert_called_once_with(
            self.mock_si, ns="pbm/2.0", path="/pbm/sdk"
        )

    def test_pbm_si(self):
        mock_get_pbm_si = MagicMock()
        with patch("salt.utils.pbm.pbm.ServiceInstance", mock_get_pbm_si):
            salt.utils.pbm.get_profile_manager(self.mock_si)
        mock_get_pbm_si.assert_called_once_with("ServiceInstance", self.mock_stub)

    def test_return_profile_manager(self):
        ret = salt.utils.pbm.get_profile_manager(self.mock_si)
        self.assertEqual(ret, self.mock_prof_mgr)

    def test_profile_manager_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        type(self.mock_content).profileManager = PropertyMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_profile_manager(self.mock_si)
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_profile_manager_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        type(self.mock_content).profileManager = PropertyMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_profile_manager(self.mock_si)
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_profile_manager_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        type(self.mock_content).profileManager = PropertyMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.get_profile_manager(self.mock_si)
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class GetPlacementSolverTestCase(TestCase):
    """Tests for salt.utils.pbm.get_placement_solver"""

    def setUp(self):
        self.mock_si = MagicMock()
        self.mock_stub = MagicMock()
        self.mock_prof_mgr = MagicMock()
        self.mock_content = MagicMock()
        self.mock_pbm_si = MagicMock(
            RetrieveContent=MagicMock(return_value=self.mock_content)
        )
        type(self.mock_content).placementSolver = PropertyMock(
            return_value=self.mock_prof_mgr
        )
        patches = (
            (
                "salt.utils.vmware.get_new_service_instance_stub",
                MagicMock(return_value=self.mock_stub),
            ),
            (
                "salt.utils.pbm.pbm.ServiceInstance",
                MagicMock(return_value=self.mock_pbm_si),
            ),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

    def tearDown(self):
        for attr in (
            "mock_si",
            "mock_stub",
            "mock_content",
            "mock_pbm_si",
            "mock_prof_mgr",
        ):
            delattr(self, attr)

    def test_get_new_service_stub(self):
        mock_get_new_service_stub = MagicMock()
        with patch(
            "salt.utils.vmware.get_new_service_instance_stub", mock_get_new_service_stub
        ):
            salt.utils.pbm.get_placement_solver(self.mock_si)
        mock_get_new_service_stub.assert_called_once_with(
            self.mock_si, ns="pbm/2.0", path="/pbm/sdk"
        )

    def test_pbm_si(self):
        mock_get_pbm_si = MagicMock()
        with patch("salt.utils.pbm.pbm.ServiceInstance", mock_get_pbm_si):
            salt.utils.pbm.get_placement_solver(self.mock_si)
        mock_get_pbm_si.assert_called_once_with("ServiceInstance", self.mock_stub)

    def test_return_profile_manager(self):
        ret = salt.utils.pbm.get_placement_solver(self.mock_si)
        self.assertEqual(ret, self.mock_prof_mgr)

    def test_placement_solver_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        type(self.mock_content).placementSolver = PropertyMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_placement_solver(self.mock_si)
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_placement_solver_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        type(self.mock_content).placementSolver = PropertyMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_placement_solver(self.mock_si)
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_placement_solver_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        type(self.mock_content).placementSolver = PropertyMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.get_placement_solver(self.mock_si)
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class GetCapabilityDefinitionsTestCase(TestCase):
    """Tests for salt.utils.pbm.get_capability_definitions"""

    def setUp(self):
        self.mock_res_type = MagicMock()
        self.mock_cap_cats = [
            MagicMock(capabilityMetadata=["fake_cap_meta1", "fake_cap_meta2"]),
            MagicMock(capabilityMetadata=["fake_cap_meta3"]),
        ]
        self.mock_prof_mgr = MagicMock(
            FetchCapabilityMetadata=MagicMock(return_value=self.mock_cap_cats)
        )
        patches = (
            (
                "salt.utils.pbm.pbm.profile.ResourceType",
                MagicMock(return_value=self.mock_res_type),
            ),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

    def tearDown(self):
        for attr in ("mock_res_type", "mock_cap_cats", "mock_prof_mgr"):
            delattr(self, attr)

    def test_get_res_type(self):
        mock_get_res_type = MagicMock()
        with patch("salt.utils.pbm.pbm.profile.ResourceType", mock_get_res_type):
            salt.utils.pbm.get_capability_definitions(self.mock_prof_mgr)
        mock_get_res_type.assert_called_once_with(
            resourceType=pbm.profile.ResourceTypeEnum.STORAGE
        )

    def test_fetch_capabilities(self):
        salt.utils.pbm.get_capability_definitions(self.mock_prof_mgr)
        self.mock_prof_mgr.FetchCapabilityMetadata.assert_called_once_with(
            self.mock_res_type
        )

    def test_fetch_capabilities_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        self.mock_prof_mgr.FetchCapabilityMetadata = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_capability_definitions(self.mock_prof_mgr)
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_fetch_capabilities_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        self.mock_prof_mgr.FetchCapabilityMetadata = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_capability_definitions(self.mock_prof_mgr)
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_fetch_capabilities_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        self.mock_prof_mgr.FetchCapabilityMetadata = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.get_capability_definitions(self.mock_prof_mgr)
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")

    def test_return_cap_definitions(self):
        ret = salt.utils.pbm.get_capability_definitions(self.mock_prof_mgr)
        self.assertEqual(ret, ["fake_cap_meta1", "fake_cap_meta2", "fake_cap_meta3"])


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class GetPoliciesByIdTestCase(TestCase):
    """Tests for salt.utils.pbm.get_policies_by_id"""

    def setUp(self):
        self.policy_ids = MagicMock()
        self.mock_policies = MagicMock()
        self.mock_prof_mgr = MagicMock(
            RetrieveContent=MagicMock(return_value=self.mock_policies)
        )

    def tearDown(self):
        for attr in ("policy_ids", "mock_policies", "mock_prof_mgr"):
            delattr(self, attr)

    def test_retrieve_policies(self):
        salt.utils.pbm.get_policies_by_id(self.mock_prof_mgr, self.policy_ids)
        self.mock_prof_mgr.RetrieveContent.assert_called_once_with(self.policy_ids)

    def test_retrieve_policies_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        self.mock_prof_mgr.RetrieveContent = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_policies_by_id(self.mock_prof_mgr, self.policy_ids)
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_retrieve_policies_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        self.mock_prof_mgr.RetrieveContent = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_policies_by_id(self.mock_prof_mgr, self.policy_ids)
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_retrieve_policies_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        self.mock_prof_mgr.RetrieveContent = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.get_policies_by_id(self.mock_prof_mgr, self.policy_ids)
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")

    def test_return_policies(self):
        ret = salt.utils.pbm.get_policies_by_id(self.mock_prof_mgr, self.policy_ids)
        self.assertEqual(ret, self.mock_policies)


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class GetStoragePoliciesTestCase(TestCase):
    """Tests for salt.utils.pbm.get_storage_policies"""

    def setUp(self):
        self.mock_res_type = MagicMock()
        self.mock_policy_ids = MagicMock()
        self.mock_prof_mgr = MagicMock(
            QueryProfile=MagicMock(return_value=self.mock_policy_ids)
        )
        # Policies
        self.mock_policies = []
        for i in range(4):
            mock_obj = MagicMock(
                resourceType=MagicMock(
                    resourceType=pbm.profile.ResourceTypeEnum.STORAGE
                )
            )
            mock_obj.name = f"fake_policy{i}"
            self.mock_policies.append(mock_obj)
        patches = (
            (
                "salt.utils.pbm.pbm.profile.ResourceType",
                MagicMock(return_value=self.mock_res_type),
            ),
            (
                "salt.utils.pbm.get_policies_by_id",
                MagicMock(return_value=self.mock_policies),
            ),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

    def tearDown(self):
        for attr in (
            "mock_res_type",
            "mock_policy_ids",
            "mock_policies",
            "mock_prof_mgr",
        ):
            delattr(self, attr)

    def test_get_res_type(self):
        mock_get_res_type = MagicMock()
        with patch("salt.utils.pbm.pbm.profile.ResourceType", mock_get_res_type):
            salt.utils.pbm.get_storage_policies(self.mock_prof_mgr)
        mock_get_res_type.assert_called_once_with(
            resourceType=pbm.profile.ResourceTypeEnum.STORAGE
        )

    def test_retrieve_policy_ids(self):
        mock_retrieve_policy_ids = MagicMock(return_value=self.mock_policy_ids)
        self.mock_prof_mgr.QueryProfile = mock_retrieve_policy_ids
        salt.utils.pbm.get_storage_policies(self.mock_prof_mgr)
        mock_retrieve_policy_ids.assert_called_once_with(self.mock_res_type)

    def test_retrieve_policy_ids_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        self.mock_prof_mgr.QueryProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_storage_policies(self.mock_prof_mgr)
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_retrieve_policy_ids_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        self.mock_prof_mgr.QueryProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_storage_policies(self.mock_prof_mgr)
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_retrieve_policy_ids_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        self.mock_prof_mgr.QueryProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.get_storage_policies(self.mock_prof_mgr)
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")

    def test_get_policies_by_id(self):
        mock_get_policies_by_id = MagicMock(return_value=self.mock_policies)
        with patch("salt.utils.pbm.get_policies_by_id", mock_get_policies_by_id):
            salt.utils.pbm.get_storage_policies(self.mock_prof_mgr)
        mock_get_policies_by_id.assert_called_once_with(
            self.mock_prof_mgr, self.mock_policy_ids
        )

    def test_return_all_policies(self):
        ret = salt.utils.pbm.get_storage_policies(
            self.mock_prof_mgr, get_all_policies=True
        )
        self.assertEqual(ret, self.mock_policies)

    def test_return_filtered_policies(self):
        ret = salt.utils.pbm.get_storage_policies(
            self.mock_prof_mgr, policy_names=["fake_policy1", "fake_policy3"]
        )
        self.assertEqual(ret, [self.mock_policies[1], self.mock_policies[3]])


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class CreateStoragePolicyTestCase(TestCase):
    """Tests for salt.utils.pbm.create_storage_policy"""

    def setUp(self):
        self.mock_policy_spec = MagicMock()
        self.mock_prof_mgr = MagicMock()

    def tearDown(self):
        for attr in ("mock_policy_spec", "mock_prof_mgr"):
            delattr(self, attr)

    def test_create_policy(self):
        salt.utils.pbm.create_storage_policy(self.mock_prof_mgr, self.mock_policy_spec)
        self.mock_prof_mgr.Create.assert_called_once_with(self.mock_policy_spec)

    def test_create_policy_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        self.mock_prof_mgr.Create = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.create_storage_policy(
                self.mock_prof_mgr, self.mock_policy_spec
            )
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_create_policy_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        self.mock_prof_mgr.Create = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.create_storage_policy(
                self.mock_prof_mgr, self.mock_policy_spec
            )
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_create_policy_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        self.mock_prof_mgr.Create = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.create_storage_policy(
                self.mock_prof_mgr, self.mock_policy_spec
            )
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class UpdateStoragePolicyTestCase(TestCase):
    """Tests for salt.utils.pbm.update_storage_policy"""

    def setUp(self):
        self.mock_policy_spec = MagicMock()
        self.mock_policy = MagicMock()
        self.mock_prof_mgr = MagicMock()

    def tearDown(self):
        for attr in ("mock_policy_spec", "mock_policy", "mock_prof_mgr"):
            delattr(self, attr)

    def test_create_policy(self):
        salt.utils.pbm.update_storage_policy(
            self.mock_prof_mgr, self.mock_policy, self.mock_policy_spec
        )
        self.mock_prof_mgr.Update.assert_called_once_with(
            self.mock_policy.profileId, self.mock_policy_spec
        )

    def test_create_policy_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        self.mock_prof_mgr.Update = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.update_storage_policy(
                self.mock_prof_mgr, self.mock_policy, self.mock_policy_spec
            )
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_create_policy_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        self.mock_prof_mgr.Update = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.update_storage_policy(
                self.mock_prof_mgr, self.mock_policy, self.mock_policy_spec
            )
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_create_policy_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        self.mock_prof_mgr.Update = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.update_storage_policy(
                self.mock_prof_mgr, self.mock_policy, self.mock_policy_spec
            )
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class GetDefaultStoragePolicyOfDatastoreTestCase(TestCase):
    """Tests for salt.utils.pbm.get_default_storage_policy_of_datastore"""

    def setUp(self):
        self.mock_ds = MagicMock(_moId="fake_ds_moid")
        self.mock_hub = MagicMock()
        self.mock_policy_id = "fake_policy_id"
        self.mock_prof_mgr = MagicMock(
            QueryDefaultRequirementProfile=MagicMock(return_value=self.mock_policy_id)
        )
        self.mock_policy_refs = [MagicMock()]
        patches = (
            (
                "salt.utils.pbm.pbm.placement.PlacementHub",
                MagicMock(return_value=self.mock_hub),
            ),
            (
                "salt.utils.pbm.get_policies_by_id",
                MagicMock(return_value=self.mock_policy_refs),
            ),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

    def tearDown(self):
        for attr in (
            "mock_ds",
            "mock_hub",
            "mock_policy_id",
            "mock_prof_mgr",
            "mock_policy_refs",
        ):
            delattr(self, attr)

    def test_get_placement_hub(self):
        mock_get_placement_hub = MagicMock()
        with patch("salt.utils.pbm.pbm.placement.PlacementHub", mock_get_placement_hub):
            salt.utils.pbm.get_default_storage_policy_of_datastore(
                self.mock_prof_mgr, self.mock_ds
            )
        mock_get_placement_hub.assert_called_once_with(
            hubId="fake_ds_moid", hubType="Datastore"
        )

    def test_query_default_requirement_profile(self):
        mock_query_prof = MagicMock(return_value=self.mock_policy_id)
        self.mock_prof_mgr.QueryDefaultRequirementProfile = mock_query_prof
        salt.utils.pbm.get_default_storage_policy_of_datastore(
            self.mock_prof_mgr, self.mock_ds
        )
        mock_query_prof.assert_called_once_with(self.mock_hub)

    def test_query_default_requirement_profile_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        self.mock_prof_mgr.QueryDefaultRequirementProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_default_storage_policy_of_datastore(
                self.mock_prof_mgr, self.mock_ds
            )
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_query_default_requirement_profile_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        self.mock_prof_mgr.QueryDefaultRequirementProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.get_default_storage_policy_of_datastore(
                self.mock_prof_mgr, self.mock_ds
            )
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_query_default_requirement_profile_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        self.mock_prof_mgr.QueryDefaultRequirementProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.get_default_storage_policy_of_datastore(
                self.mock_prof_mgr, self.mock_ds
            )
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")

    def test_get_policies_by_id(self):
        mock_get_policies_by_id = MagicMock()
        with patch("salt.utils.pbm.get_policies_by_id", mock_get_policies_by_id):
            salt.utils.pbm.get_default_storage_policy_of_datastore(
                self.mock_prof_mgr, self.mock_ds
            )
        mock_get_policies_by_id.assert_called_once_with(
            self.mock_prof_mgr, [self.mock_policy_id]
        )

    def test_no_policy_refs(self):
        mock_get_policies_by_id = MagicMock()
        with patch("salt.utils.pbm.get_policies_by_id", MagicMock(return_value=None)):
            with self.assertRaises(VMwareObjectRetrievalError) as excinfo:
                salt.utils.pbm.get_default_storage_policy_of_datastore(
                    self.mock_prof_mgr, self.mock_ds
                )
        self.assertEqual(
            excinfo.exception.strerror,
            "Storage policy with id 'fake_policy_id' was not found",
        )

    def test_return_policy_ref(self):
        mock_get_policies_by_id = MagicMock()
        ret = salt.utils.pbm.get_default_storage_policy_of_datastore(
            self.mock_prof_mgr, self.mock_ds
        )
        self.assertEqual(ret, self.mock_policy_refs[0])


@pytest.mark.skipif(not HAS_PYVMOMI, reason="The 'pyvmomi' library is missing")
class AssignDefaultStoragePolicyToDatastoreTestCase(TestCase):
    """Tests for salt.utils.pbm.assign_default_storage_policy_to_datastore"""

    def setUp(self):
        self.mock_ds = MagicMock(_moId="fake_ds_moid")
        self.mock_policy = MagicMock()
        self.mock_hub = MagicMock()
        self.mock_prof_mgr = MagicMock()
        patches = (
            (
                "salt.utils.pbm.pbm.placement.PlacementHub",
                MagicMock(return_value=self.mock_hub),
            ),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

    def tearDown(self):
        for attr in ("mock_ds", "mock_hub", "mock_policy", "mock_prof_mgr"):
            delattr(self, attr)

    def test_get_placement_hub(self):
        mock_get_placement_hub = MagicMock()
        with patch("salt.utils.pbm.pbm.placement.PlacementHub", mock_get_placement_hub):
            salt.utils.pbm.assign_default_storage_policy_to_datastore(
                self.mock_prof_mgr, self.mock_policy, self.mock_ds
            )
        mock_get_placement_hub.assert_called_once_with(
            hubId="fake_ds_moid", hubType="Datastore"
        )

    def test_assign_default_requirement_profile(self):
        mock_assign_prof = MagicMock()
        self.mock_prof_mgr.AssignDefaultRequirementProfile = mock_assign_prof
        salt.utils.pbm.assign_default_storage_policy_to_datastore(
            self.mock_prof_mgr, self.mock_policy, self.mock_ds
        )
        mock_assign_prof.assert_called_once_with(
            self.mock_policy.profileId, [self.mock_hub]
        )

    def test_assign_default_requirement_profile_raises_no_permissions(self):
        exc = vim.fault.NoPermission()
        exc.privilegeId = "Fake privilege"
        self.mock_prof_mgr.AssignDefaultRequirementProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.assign_default_storage_policy_to_datastore(
                self.mock_prof_mgr, self.mock_policy, self.mock_ds
            )
        self.assertEqual(
            excinfo.exception.strerror,
            "Not enough permissions. Required privilege: Fake privilege",
        )

    def test_assign_default_requirement_profile_raises_vim_fault(self):
        exc = vim.fault.VimFault()
        exc.msg = "VimFault msg"
        self.mock_prof_mgr.AssignDefaultRequirementProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareApiError) as excinfo:
            salt.utils.pbm.assign_default_storage_policy_to_datastore(
                self.mock_prof_mgr, self.mock_policy, self.mock_ds
            )
        self.assertEqual(excinfo.exception.strerror, "VimFault msg")

    def test_assign_default_requirement_profile_raises_runtime_fault(self):
        exc = vmodl.RuntimeFault()
        exc.msg = "RuntimeFault msg"
        self.mock_prof_mgr.AssignDefaultRequirementProfile = MagicMock(side_effect=exc)
        with self.assertRaises(VMwareRuntimeError) as excinfo:
            salt.utils.pbm.assign_default_storage_policy_to_datastore(
                self.mock_prof_mgr, self.mock_policy, self.mock_ds
            )
        self.assertEqual(excinfo.exception.strerror, "RuntimeFault msg")



import pytest

import salt.modules.rebootmgr as rebootmgr
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {rebootmgr: {"__salt__": {}, "__utils__": {}}}


def test_version():
    """
    Test rebootmgr.version without parameters
    """
    version = "rebootmgrctl (rebootmgr) 1.3"
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": version, "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.version() == "1.3"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "--version"])


def test_is_active():
    """
    Test rebootmgr.is_active without parameters
    """
    salt_mock = {"cmd.run_all": MagicMock(return_value={"stdout": None, "retcode": 0})}
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.is_active()
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "is_active", "--quiet"]
        )


def test_reboot():
    """
    Test rebootmgr.reboot without parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.reboot() == "output"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "reboot"])


def test_reboot_order():
    """
    Test rebootmgr.reboot with order parameter
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.reboot("now") == "output"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "reboot", "now"])


def test_reboot_invalid():
    """
    Test rebootmgr.reboot with invalid parameter
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        with pytest.raises(CommandExecutionError):
            rebootmgr.reboot("invalid")


def test_cancel():
    """
    Test rebootmgr.cancel without parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.cancel() == "output"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "cancel"])


def test_status():
    """
    Test rebootmgr.status without parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        # 0 - No reboot requested
        assert rebootmgr.status() == 0
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "status", "--quiet"]
        )


def test_set_strategy_default():
    """
    Test rebootmgr.set_strategy without parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.set_strategy() == "output"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "set-strategy"])


def test_set_strategy():
    """
    Test rebootmgr.set_strategy with strategy parameter
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.set_strategy("best-effort") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "set-strategy", "best-effort"]
        )


def test_set_strategy_invalid():
    """
    Test rebootmgr.strategy with invalid parameter
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        with pytest.raises(CommandExecutionError):
            rebootmgr.set_strategy("invalid")


def test_get_strategy():
    """
    Test rebootmgr.get_strategy without parameters
    """
    strategy = "Reboot strategy: best-effort"
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": strategy, "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.get_strategy() == "best-effort"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "get-strategy"])


def test_set_window():
    """
    Test rebootmgr.set_window with parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.set_window("Thu,Fri 2020-*-1,5 11:12:13", "1h") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "set-window", "Thu,Fri 2020-*-1,5 11:12:13", "1h"]
        )


def test_get_window():
    """
    Test rebootmgr.get_window without parameters
    """
    window = "Maintenance window is set to *-*-* 03:30:00, lasting 01h30m."
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": window, "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.get_window() == {
            "time": "*-*-* 03:30:00",
            "duration": "01h30m",
        }
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "get-window"])


def test_set_group():
    """
    Test rebootmgr.set_group with parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.set_group("group1") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "set-group", "group1"]
        )


def test_get_group():
    """
    Test rebootmgr.get_group without parameters
    """
    group = "Etcd lock group is set to group1"
    salt_mock = {"cmd.run_all": MagicMock(return_value={"stdout": group, "retcode": 0})}
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.get_group() == "group1"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "get-group"])


def test_set_max():
    """
    Test rebootmgr.set_max with default parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.set_max(10) == "output"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "set-max", 10])


def test_set_max_group():
    """
    Test rebootmgr.set_max with group parameter
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.set_max(10, "group1") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "set-max", "--group", "group1", 10]
        )


def test_lock():
    """
    Test rebootmgr.lock without parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.lock() == "output"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "lock"])


def test_lock_machine_id():
    """
    Test rebootmgr.lock with machine_id parameter
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.lock("machine-id") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "lock", "machine-id"]
        )


def test_lock_machine_id_group():
    """
    Test rebootmgr.lock with machine_id and group parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.lock("machine-id", "group1") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "lock", "--group", "group1", "machine-id"]
        )


def test_unlock():
    """
    Test rebootmgr.unlock without parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.unlock() == "output"
        salt_mock["cmd.run_all"].assert_called_with(["rebootmgrctl", "unlock"])


def test_unlock_machine_id():
    """
    Test rebootmgr.unlock with machine_id parameter
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.unlock("machine-id") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "unlock", "machine-id"]
        )


def test_unlock_machine_id_group():
    """
    Test rebootmgr.unlock with machine_id and group parameters
    """
    salt_mock = {
        "cmd.run_all": MagicMock(return_value={"stdout": "output", "retcode": 0})
    }
    with patch.dict(rebootmgr.__salt__, salt_mock):
        assert rebootmgr.unlock("machine-id", "group1") == "output"
        salt_mock["cmd.run_all"].assert_called_with(
            ["rebootmgrctl", "unlock", "--group", "group1", "machine-id"]
        )



"""
Tests for the SVN state
"""

import logging

import pytest

import salt.utils.platform
from tests.support.helpers import SKIP_INITIAL_PHOTONOS_FAILURES

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.requires_network,
    pytest.mark.skip_if_binaries_missing("svn"),
]


@pytest.fixture
def repo_target(tmp_path_world_rw):
    return tmp_path_world_rw / "svn-repo-checkout"


@pytest.fixture
def repo_revision():
    return "1456987"


@pytest.fixture
def repo_url():
    return "http://svn.apache.org/repos/asf/httpd/httpd/trunk/test/"


@pytest.fixture(scope="module")
def svn(states):
    return states.svn


@pytest.fixture(scope="module")
def svn_mod(modules):
    return modules.svn


@pytest.fixture(scope="module")
def account():
    with pytest.helpers.create_account(create_group=True) as _account:
        yield _account


@pytest.mark.slow_test
def test_latest(svn, repo_url, repo_revision, repo_target):
    """
    svn.latest
    """
    ret = svn.latest(name=repo_url, rev=repo_revision, target=str(repo_target))
    assert ret.result is True
    assert repo_target.joinpath(".svn").is_dir()
    assert ret.changes
    assert "new" in ret.changes
    assert ret.changes["new"] == repo_url
    assert "revision" in ret.changes
    assert ret.changes["revision"] == repo_revision


@pytest.mark.slow_test
def test_latest_failure(svn, repo_revision, repo_target):
    """
    svn.latest
    """
    ret = svn.latest(
        name="https://youSpelledApacheWrong.com/repo/asf/httpd/trunk/",
        rev=repo_revision,
        target=str(repo_target),
    )
    assert ret.result is False
    assert not repo_target.joinpath(".svn").is_dir()


@SKIP_INITIAL_PHOTONOS_FAILURES
@pytest.mark.slow_test
@pytest.mark.destructive_test
@pytest.mark.skip_if_not_root
def test_latest_user(svn, repo_url, repo_revision, repo_target, account):
    """
    svn.latest
    """
    ret = svn.latest(
        name=repo_url, rev=repo_revision, target=str(repo_target), user=account.username
    )
    assert ret.result is True
    assert repo_target.joinpath(".svn").is_dir()
    assert ret.changes
    assert "new" in ret.changes
    assert ret.changes["new"] == repo_url
    assert "revision" in ret.changes
    assert ret.changes["revision"] == repo_revision

    # Make sure that the files in the cloned repo are owned by the account that did the checkout
    for entry in repo_target.iterdir():
        entry_stat = entry.stat()
        assert entry_stat.st_uid == account.info.uid
        try:
            assert entry_stat.st_gid == account.info.gid
        except AssertionError:
            if not salt.utils.platform.is_darwin():
                raise
            pytest.xfail("The 'cmd' module does not change to the user group on Darwin")


@pytest.mark.slow_test
def test_latest_empty_dir(svn, repo_url, repo_revision, repo_target):
    """
    svn.latest
    """
    repo_target.mkdir()
    ret = svn.latest(name=repo_url, rev=repo_revision, target=str(repo_target))
    assert ret.result is True
    assert repo_target.joinpath(".svn").is_dir()


def no_test_latest_existing_repo(svn, svn_mod, repo_url, repo_revision, repo_target):
    """
    svn.latest against existing repository
    """
    current_rev = "1442865"
    cwd = str(repo_target.parent)
    basename = repo_target.name
    opts = ("-r", current_rev)

    assert svn_mod.checkout(cwd, repo_url, basename, None, None, opts)

    ret = svn.latest(name=repo_url, rev=repo_revision, target=str(repo_target))
    assert ret.result is True
    assert ret.changes
    assert "revision" in ret.changes
    assert ret.changes["revision"] == f"{current_rev} => {repo_revision}"
    assert repo_target.joinpath(".svn").is_dir()


def no_test_latest_existing_repo_no_rev_change(
    svn, svn_mod, repo_url, repo_revision, repo_target
):
    """
    svn.latest against existing repository
    """
    current_rev = repo_revision
    cwd = str(repo_target.parent)
    basename = repo_target.name
    opts = ("-r", current_rev)
    assert svn_mod.checkout(cwd, repo_url, basename, None, None, opts)

    ret = svn.latest(name=repo_url, rev=repo_revision, target=str(repo_target))
    assert ret.result is True
    assert not ret.changes
    assert isinstance(ret.changes, dict)
    assert repo_target.joinpath(".svn").is_dir()



import subprocess
import textwrap

import pytest


@pytest.fixture
def python_script_bin(install_salt):
    return install_salt.binary_paths["python"]


@pytest.fixture
def check_python_file(tmp_path):
    script_path = tmp_path / "check_python.py"
    script_path.write_text(
        textwrap.dedent(
            """
        import sys

        import salt.utils.data

        user_arg = sys.argv

        if user_arg[1] == "raise":
            raise Exception("test")

        if salt.utils.data.is_true(user_arg[1]):
            sys.exit(0)
        else:
            sys.exit(1)
        """
        )
    )
    return script_path


@pytest.mark.parametrize("exp_ret,user_arg", [(1, "false"), (0, "true")])
def test_python_script(
    install_salt, exp_ret, user_arg, python_script_bin, check_python_file
):
    ret = install_salt.proc.run(
        *(
            python_script_bin
            + [
                str(check_python_file),
                user_arg,
            ]
        ),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        universal_newlines=True,
    )

    assert ret.returncode == exp_ret, ret.stderr


def test_python_script_exception(install_salt, python_script_bin, check_python_file):
    ret = install_salt.proc.run(
        *(
            python_script_bin
            + [
                str(check_python_file),
                "raise",
            ]
        ),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        universal_newlines=True,
    )
    assert "Exception: test" in ret.stderr



"""
Validate Cache class methods
"""

import pytest

import salt.cache
import salt.payload
from tests.support.mock import patch


@pytest.fixture
def opts():
    return {
        "cache": "fake_driver",
        "memcache_expire_seconds": 10,
        "memcache_max_items": 3,
        "memcache_full_cleanup": False,
        "memcache_debug": False,
    }


@pytest.fixture
def cache(opts):
    salt.cache.MemCache.data = {}
    return salt.cache.factory(opts)


def test_fetch(cache):
    with patch("salt.cache.Cache.fetch", return_value="fake_data") as cache_fetch_mock:
        with patch("salt.loader.cache", return_value={}):
            # Fetch value, it will be kept in cache.
            with patch("time.time", return_value=0):
                ret = cache.fetch("bank", "key")
            assert ret == "fake_data"
            assert salt.cache.MemCache.data == {
                "fake_driver": {("bank", "key"): [0, "fake_data"]}
            }
            cache_fetch_mock.assert_called_once_with("bank", "key")
            cache_fetch_mock.reset_mock()

            # Fetch again, cached value is used, time updated.
            with patch("time.time", return_value=1):
                ret = cache.fetch("bank", "key")
            assert ret == "fake_data"
            assert salt.cache.MemCache.data == {
                "fake_driver": {("bank", "key"): [1, "fake_data"]}
            }
            cache_fetch_mock.assert_not_called()

            # Fetch after expire
            with patch("time.time", return_value=12):
                ret = cache.fetch("bank", "key")
            assert ret == "fake_data"
            assert salt.cache.MemCache.data == {
                "fake_driver": {("bank", "key"): [12, "fake_data"]}
            }
            cache_fetch_mock.assert_called_once_with("bank", "key")
            cache_fetch_mock.reset_mock()


def test_store(cache):
    with patch("salt.cache.Cache.store") as cache_store_mock:
        with patch("salt.loader.cache", return_value={}):
            # Fetch value, it will be kept in cache.
            with patch("time.time", return_value=0):
                cache.store("bank", "key", "fake_data")
            assert salt.cache.MemCache.data == {
                "fake_driver": {("bank", "key"): [0, "fake_data"]}
            }
            cache_store_mock.assert_called_once_with("bank", "key", "fake_data")
            cache_store_mock.reset_mock()

            # Store another value.
            with patch("time.time", return_value=1):
                cache.store("bank", "key2", "fake_data2")
            assert salt.cache.MemCache.data == {
                "fake_driver": {
                    ("bank", "key"): [0, "fake_data"],
                    ("bank", "key2"): [1, "fake_data2"],
                }
            }
            cache_store_mock.assert_called_once_with("bank", "key2", "fake_data2")


def test_flush(cache):
    with patch("salt.cache.Cache.flush") as cache_flush_mock:
        with patch("salt.cache.Cache.store"):
            with patch("salt.loader.cache", return_value={}):
                # Flush non-existing bank
                cache.flush("bank")
                assert salt.cache.MemCache.data == {"fake_driver": {}}
                cache_flush_mock.assert_called_once_with("bank", None)
                cache_flush_mock.reset_mock()
                # Flush non-existing key
                cache.flush("bank", "key")
                assert salt.cache.MemCache.data == {"fake_driver": {}}
                cache_flush_mock.assert_called_once_with("bank", "key")
                cache_flush_mock.reset_mock()
                # Flush existing key
                with patch("time.time", return_value=0):
                    cache.store("bank", "key", "fake_data")
                assert salt.cache.MemCache.data["fake_driver"][("bank", "key")] == [
                    0,
                    "fake_data",
                ]
                assert salt.cache.MemCache.data == {
                    "fake_driver": {("bank", "key"): [0, "fake_data"]}
                }
                cache.flush("bank", "key")
                assert salt.cache.MemCache.data == {"fake_driver": {}}
                cache_flush_mock.assert_called_once_with("bank", "key")
                cache_flush_mock.reset_mock()


def test_max_items(cache):
    with patch("salt.cache.Cache.store"):
        with patch("salt.loader.cache", return_value={}):
            # Put MAX=3 values
            with patch("time.time", return_value=0):
                cache.store("bank1", "key1", "fake_data11")
            with patch("time.time", return_value=1):
                cache.store("bank1", "key2", "fake_data12")
            with patch("time.time", return_value=2):
                cache.store("bank2", "key1", "fake_data21")
            assert salt.cache.MemCache.data["fake_driver"] == {
                ("bank1", "key1"): [0, "fake_data11"],
                ("bank1", "key2"): [1, "fake_data12"],
                ("bank2", "key1"): [2, "fake_data21"],
            }
            # Put one more and check the oldest was removed
            with patch("time.time", return_value=3):
                cache.store("bank2", "key2", "fake_data22")
            assert salt.cache.MemCache.data["fake_driver"] == {
                ("bank1", "key2"): [1, "fake_data12"],
                ("bank2", "key1"): [2, "fake_data21"],
                ("bank2", "key2"): [3, "fake_data22"],
            }


def test_full_cleanup(cache):
    with patch("salt.cache.Cache.store"):
        with patch("salt.loader.cache", return_value={}):
            # Enable full cleanup
            cache.cleanup = True
            # Put MAX=3 values
            with patch("time.time", return_value=0):
                cache.store("bank1", "key1", "fake_data11")
            with patch("time.time", return_value=1):
                cache.store("bank1", "key2", "fake_data12")
            with patch("time.time", return_value=2):
                cache.store("bank2", "key1", "fake_data21")
            assert salt.cache.MemCache.data["fake_driver"] == {
                ("bank1", "key1"): [0, "fake_data11"],
                ("bank1", "key2"): [1, "fake_data12"],
                ("bank2", "key1"): [2, "fake_data21"],
            }
            # Put one more and check all expired was removed
            with patch("time.time", return_value=12):
                cache.store("bank2", "key2", "fake_data22")
            assert salt.cache.MemCache.data["fake_driver"] == {
                ("bank2", "key1"): [2, "fake_data21"],
                ("bank2", "key2"): [12, "fake_data22"],
            }


def test_fetch_debug(cache, opts):
    with patch("salt.cache.Cache.fetch", return_value="fake_data"):
        with patch("salt.loader.cache", return_value={}):
            # Recreate cache with debug enabled
            opts["memcache_debug"] = True
            cache = salt.cache.factory(opts)

            # Fetch 2 values (no cache hit)
            with patch("time.time", return_value=0):
                ret = cache.fetch("bank", "key1")
            with patch("time.time", return_value=1):
                ret = cache.fetch("bank", "key2")
            # Fetch 3 times (cache hit)
            with patch("time.time", return_value=2):
                ret = cache.fetch("bank", "key2")
            with patch("time.time", return_value=3):
                ret = cache.fetch("bank", "key1")
            with patch("time.time", return_value=4):
                ret = cache.fetch("bank", "key1")
            # Fetch an expired value (no cache hit)
            with patch("time.time", return_value=13):
                ret = cache.fetch("bank", "key2")

            # Check debug data
            assert cache.call == 6
            assert cache.hit == 3
